# coding: utf-8
"""

"""
import csv
import io
import json
import re
import typing

import math

import flask
import flask_login
import itsdangerous
from flask_babel import _

from .. import frontend
from ... import logic
from ... import models
from ...logic import object_log, comments, errors
from ...logic.actions import get_action
from ...logic.action_types import get_action_type
from ...logic.action_permissions import get_user_action_permissions
from ...logic.caching import cache_per_request
from ...logic.object_permissions import get_user_object_permissions
from ...logic.fed_logs import get_fed_object_log_entries_for_object
from ...logic.object_relationships import get_workflow_references
from ...logic.users import get_user, get_users, User
from ...logic.settings import get_user_settings
from ...logic.objects import get_object
from ...logic.projects import get_project, Project
from ...logic.locations import get_location, get_object_location_assignment, get_object_location_assignments, assign_location_to_object
from ...logic.locations import Location, get_location_capacities, get_assigned_object_count_by_action_types
from ...logic.location_permissions import get_locations_with_user_permissions
from ...logic.languages import get_language_by_lang_code, get_language, get_languages, Language
from ...logic.components import get_component
from ...logic.shares import get_shares_for_object
from ...logic.notebook_templates import get_notebook_templates
from ...logic.utils import get_translated_text, get_data_and_schema_by_id_path
from .forms import ObjectForm, CommentForm, FileForm, FileInformationForm, FileHidingForm, ObjectLocationAssignmentForm, ExternalLinkForm, ObjectPublicationForm, GenerateLabelsForm
from ...utils import object_permissions_required
from ..utils import generate_qrcode, get_user_if_exists, get_locations_form_data
from ..labels import create_labels, create_multiple_labels, PAGE_SIZES, PAGE_SIZE_KEYS, DEFAULT_PAPER_FORMAT, HORIZONTAL_LABEL_MARGIN, VERTICAL_LABEL_MARGIN
from .. import pdfexport
from ..utils import check_current_user_is_not_readonly, get_location_name
from .permissions import on_unauthorized, get_object_if_current_user_has_read_permissions, get_fed_object_if_current_user_has_read_permissions
from .object_form import show_object_form, get_object_form_template_kwargs
from ...utils import FlaskResponseT
from ...models import Permissions, ObjectLogEntryType
from ...models.file_log import FileLogEntryType


def build_object_location_assignment_confirmation_url(object_location_assignment_id: int) -> str:
    confirmation_url = flask.url_for(
        'frontend.accept_responsibility_for_object',
        t=logic.security_tokens.generate_token(
            object_location_assignment_id,
            salt='confirm_responsibility',
            secret_key=flask.current_app.config['SECRET_KEY']
        ),
        _external=True
    )
    return confirmation_url


def build_object_location_assignment_declination_url(object_location_assignment_id: int) -> str:
    declination_url = flask.url_for(
        'frontend.decline_responsibility_for_object',
        t=logic.security_tokens.generate_token(
            object_location_assignment_id,
            salt='decline_responsibility',
            secret_key=flask.current_app.config['SECRET_KEY']
        ),
        _external=True
    )
    return declination_url


def get_project_if_it_exists(project_id: int) -> typing.Optional[Project]:
    try:
        return get_project(project_id)
    except errors.ProjectDoesNotExistError:
        return None


@frontend.route('/objects/<int:object_id>', methods=['GET', 'POST'])
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object(object_id: int) -> FlaskResponseT:
    object = get_object(object_id=object_id)
    user_permissions = get_user_object_permissions(object_id=object_id, user_id=flask_login.current_user.id)
    user_may_edit = Permissions.WRITE in user_permissions and object.fed_object_id is None and object.action_id is not None
    user_may_assign_location = user_may_edit
    user_may_link_publication = user_may_edit
    user_may_comment = user_may_edit
    user_may_upload_files = user_may_edit
    if object.eln_import_id is not None:
        user_may_edit = False
    user_may_grant = Permissions.GRANT in user_permissions

    mode = flask.request.args.get('mode', '')
    if not user_may_edit and mode in {'edit', 'upgrade'}:
        if object.fed_object_id is not None or object.eln_import_id is not None:
            flask.flash(_('Editing imported objects is not yet supported.'), 'error')
        return flask.abort(403)
    if not flask.current_app.config['DISABLE_INLINE_EDIT']:
        if not user_may_edit and mode == 'inline_edit':
            return flask.abort(403)
    if mode in {'edit', 'upgrade'} or flask.request.method == 'POST':
        check_current_user_is_not_readonly()
        action = get_action(object.action_id) if object.action_id else None
        if action is None:
            return flask.abort(400)
        should_upgrade_schema = mode == 'upgrade'
        if should_upgrade_schema and (not action.schema or action.schema == object.schema):
            flask.flash(_('The schema for this object cannot be updated.'), 'error')
            return flask.redirect(flask.url_for('.object', object_id=object_id))
        return show_object_form(
            object=object,
            action=action,
            should_upgrade_schema=should_upgrade_schema
        )

    template_kwargs: typing.Dict[str, typing.Any] = {}

    # languages
    english = get_language(Language.ENGLISH)
    all_languages = get_languages()
    languages_by_lang_code = {
        language.lang_code: language
        for language in all_languages
    }
    if object.data:
        object_languages = [
            languages_by_lang_code.get(lang_code)
            for lang_code in logic.languages.get_languages_in_object_data(object.data)
        ]
    else:
        object_languages = []
    metadata_language = flask.request.args.get('language', None)
    if metadata_language not in languages_by_lang_code:
        metadata_language = None
    template_kwargs.update({
        "metadata_language": metadata_language,
        "languages": object_languages,  # Will be overwritten in inline edit
        "object_languages": object_languages,
        "all_languages": all_languages,
        "SUPPORTED_LOCALES": logic.locale.SUPPORTED_LOCALES,
        "ENGLISH": english,
        "eln_import_url": logic.eln_import.get_eln_import_object_url(object_id)
    })

    actions_by_id = {}

    # basic object and action information
    if object.action_id is not None:
        action = get_action(object.action_id)
        actions_by_id[action.id] = action
        action_type = get_action_type(action.type_id) if action.type_id is not None else None
        instrument = action.instrument
        object_type = get_translated_text(action_type.object_name) if action_type else None
        new_schema_available = action.schema is not None and action.schema != object.schema
        user_may_use_as_template = Permissions.READ in get_user_action_permissions(object.action_id, user_id=flask_login.current_user.id)
        if logic.action_types.is_usable_in_action_types_table_empty() and not flask.current_app.config['DISABLE_USE_IN_MEASUREMENT'] and action_type and action_type.id == models.ActionType.SAMPLE_CREATION:
            usable_in_action_types = [logic.action_types.get_action_type(models.ActionType.MEASUREMENT)]
        else:
            usable_in_action_types = action_type.usable_in_action_types if action_type is not None else []
    else:
        action = None
        action_type = None
        instrument = None
        object_type = None
        new_schema_available = False
        user_may_use_as_template = False
        usable_in_action_types = None
    template_kwargs.update({
        "object": object,
        "object_id": object_id,
        "version_id": object.version_id,
        "object_type": object_type,
        "action": action,
        "action_type": action_type,
        "usable_in_action_types": usable_in_action_types,
        "instrument": instrument,
        "schema": object.schema,
        "data": object.data,
        "name": object.name,
        "component": object.component,
        "fed_object_id": object.fed_object_id,
        "fed_version_id": object.fed_version_id,
        "new_schema_available": new_schema_available,
        "user_may_edit": user_may_edit,
        "user_may_grant": user_may_grant,
        "user_may_use_as_template": user_may_use_as_template,
        "restore_form": None,
    })

    # user settings
    user_settings = get_user_settings(flask_login.current_user.id)
    template_kwargs.update({
        "show_object_type_and_id_on_object_page_text": user_settings["SHOW_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE"],
        "show_object_title": user_settings["SHOW_OBJECT_TITLE"],
        "workflow_view_modals": flask.current_app.config['WORKFLOW_VIEW_MODALS'] if user_settings['WORKFLOW_VIEW_MODALS'] is None else user_settings['WORKFLOW_VIEW_MODALS'],
        "workflow_view_collapsed": flask.current_app.config['WORKFLOW_VIEW_COLLAPSED'] if user_settings['WORKFLOW_VIEW_COLLAPSED'] is None else user_settings['WORKFLOW_VIEW_COLLAPSED'],
    })

    # QR code
    object_url = flask.url_for('.object', object_id=object_id, _external=True)
    object_qrcode = generate_qrcode(object_url, should_cache=True)
    template_kwargs.update({
        "object_qrcode": object_qrcode,
        "object_url": object_url,
    })

    # labels
    template_kwargs.update({
        "PAGE_SIZES": PAGE_SIZES,
        "HORIZONTAL_LABEL_MARGIN": HORIZONTAL_LABEL_MARGIN,
        "VERTICAL_LABEL_MARGIN": VERTICAL_LABEL_MARGIN,
    })

    # dataverse export
    dataverse_enabled = bool(flask.current_app.config['DATAVERSE_URL'])
    dataverse_export_task_created = False
    dataverse_url = None
    show_dataverse_export = False
    if dataverse_enabled and object.eln_import_id is None:
        export_state = logic.dataverse_export.get_dataverse_export_state(object.id)
        if export_state == models.DataverseExportStatus.EXPORT_FINISHED:
            dataverse_url = logic.dataverse_export.get_dataverse_url(object.id)
        elif export_state == models.DataverseExportStatus.TASK_CREATED:
            dataverse_export_task_created = True
        show_dataverse_export = user_may_grant and not dataverse_url
    template_kwargs.update({
        "show_dataverse_export": show_dataverse_export,
        "dataverse_url": dataverse_url,
        "dataverse_export_task_created": dataverse_export_task_created
    })

    # scicat export
    scicat_enabled = bool(flask.current_app.config['SCICAT_API_URL']) and bool(flask.current_app.config['SCICAT_FRONTEND_URL'])
    if scicat_enabled and object.eln_import_id is None:
        scicat_url = logic.scicat_export.get_scicat_url(object.id)
        show_scicat_export = user_may_grant and not scicat_url and action_type is not None and action_type.scicat_export_type is not None
    else:
        scicat_url = None
        show_scicat_export = False
    template_kwargs.update({
        "show_scicat_export": show_scicat_export,
        "scicat_url": scicat_url,
    })

    # download service
    download_service_enabled = bool(flask.current_app.config['DOWNLOAD_SERVICE_URL'])
    download_service_enabled = download_service_enabled and (flask.current_app.config['DOWNLOAD_SERVICE_SECRET'])
    template_kwargs.update({
        "show_download_service": download_service_enabled,
    })

    if flask_login.current_user.is_authenticated:
        favorite_actions_by_action_type_id = {}

        if usable_in_action_types:
            all_favorite_action_ids = logic.favorites.get_user_favorite_action_ids(flask_login.current_user.id)
            for usable_action_type in usable_in_action_types:
                all_action_type_actions = logic.actions.get_actions(action_type_id=usable_action_type.id)

                favorite_actions_by_action_type_id[usable_action_type.id] = [
                    action
                    for action in all_action_type_actions
                    if action.id in all_favorite_action_ids and not action.is_hidden
                ]

                favorite_actions_by_action_type_id[usable_action_type.id].sort(
                    key=lambda action: (
                        action.user.name.lower() if action.user and action.user.name else '',
                        get_translated_text(action.instrument.name).lower() if action.instrument else '',
                        get_translated_text(action.name).lower()
                    )
                )
    else:
        favorite_actions_by_action_type_id = None

    if logic.action_types.is_usable_in_action_types_table_empty() and not flask.current_app.config['DISABLE_USE_IN_MEASUREMENT']:
        action_type_name_by_action_type_id = {
            models.ActionType.MEASUREMENT: get_translated_text(
                logic.action_types.get_action_type(models.ActionType.MEASUREMENT).name, default="Unnamed Action Type"
            )
        }
    elif usable_in_action_types:
        action_type_name_by_action_type_id = {
            action_type.id: get_translated_text(logic.action_types.get_action_type(action_type.id).name, default="Unnamed Action Type")
            for action_type in usable_in_action_types}
    else:
        action_type_name_by_action_type_id = {}

    template_kwargs.update({
        "favorite_actions_by_action_type_id": favorite_actions_by_action_type_id,
        "action_type_name_by_action_type_id": action_type_name_by_action_type_id
    })

    # notebook templates
    if object.schema is not None and object.data is not None:
        notebook_templates = get_notebook_templates(
            object_id=object.id,
            data=object.data,
            schema=object.schema,
            user_id=flask_login.current_user.id
        )
    else:
        notebook_templates = []
    template_kwargs.update({
        "notebook_templates": notebook_templates,
    })

    # linked project group
    linked_project = logic.projects.get_project_linked_to_object(object_id)
    template_kwargs.update({
        "project": linked_project,
    })

    # files
    template_kwargs.update({
        "files": logic.files.get_files_for_object(object_id),
        "file_source_instrument_exists": False,
        "file_source_jupyterhub_exists": False,
        "file_form": FileForm(),
        "edit_external_link_file": flask.request.args.get('edit_invalid_link_file', None),
        "edit_external_link_error": flask.request.args.get('edit_invalid_link_error', None),
        "external_link_form": ExternalLinkForm(),
        "external_link_error": flask.request.args.get('invalid_link_error', None),
        "external_link_errors": {
            '0': _('Please enter a valid URL.'),
            '1': _('The URL you have entered exceeds the length limit.'),
            '2': _('The IP address you entered is invalid.'),
            '3': _('The port number you entered is invalid.')
        },
        "FileLogEntryType": FileLogEntryType,
        "file_information_form": FileInformationForm(),
        "file_hiding_form": FileHidingForm(),
    })

    # mobile file upload
    if user_may_upload_files:
        serializer = itsdangerous.URLSafeTimedSerializer(flask.current_app.config['SECRET_KEY'], salt='mobile-upload')
        token = serializer.dumps([flask_login.current_user.id, object_id])
        mobile_upload_url = flask.url_for('.mobile_file_upload', object_id=object_id, token=token, _external=True)
        mobile_upload_qrcode = generate_qrcode(mobile_upload_url, should_cache=False)
    else:
        mobile_upload_url = None
        mobile_upload_qrcode = None
    template_kwargs.update({
        "user_may_upload_files": user_may_upload_files,
        "mobile_upload_url": mobile_upload_url,
        "mobile_upload_qrcode": mobile_upload_qrcode,
    })

    # object location and/or responsible user assignments
    object_location_assignments = get_object_location_assignments(object_id)
    if user_may_assign_location:
        location_form = ObjectLocationAssignmentForm()

        def location_filter(location: Location, action_type_id: typing.Optional[int] = action_type.id if action_type is not None else None) -> bool:
            if not location.enable_object_assignments:
                return False
            if not location.type.enable_object_assignments:
                return False
            if location.type.enable_capacities:
                if action_type_id is None:
                    return False
                capacity = get_location_capacities(location.id).get(action_type_id, 0)
                if capacity is not None:
                    if capacity == 0:
                        return False
                    num_stored_objects = get_assigned_object_count_by_action_types(location.id, ignored_object_ids=[object_id]).get(action_type_id, 0)
                    if num_stored_objects + 1 > capacity:
                        return False
            return True

        all_choices, choices = get_locations_form_data(filter=location_filter)
        location_form.location.all_choices = all_choices
        location_form.location.choices = choices
        possible_responsible_users: typing.List[typing.Tuple[str, typing.Optional[User]]] = [('-1', None)]
        user_is_fed = {}
        for user in get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN']):
            possible_responsible_users.append((str(user.id), user))
            user_is_fed[str(user.id)] = user.fed_id is not None
        location_form.responsible_user.choices = possible_responsible_users
    else:
        location_form = None
        user_is_fed = None
    template_kwargs.update({
        "user_may_assign_location": user_may_assign_location,
        "object_location_assignments": object_location_assignments,
        "location_form": location_form,
        "user_is_fed": user_is_fed,
    })

    # publications
    object_publications = logic.publications.get_publications_for_object(object_id=object.id)
    if user_may_link_publication:
        publication_form = ObjectPublicationForm()
    else:
        publication_form = None
    template_kwargs.update({
        "user_may_link_publication": user_may_link_publication,
        "publication_form": publication_form,
        "object_publications": object_publications,
    })

    # comments
    template_kwargs.update({
        "user_may_comment": user_may_comment,
        "comments": comments.get_comments_for_object(object_id),
        "comment_form": CommentForm(),
    })

    # activity log
    object_log_entries = object_log.get_object_log_entries(
        object_id=object_id,
        user_id=flask_login.current_user.id
    )
    template_kwargs.update({
        "object_log_entries": object_log_entries,
        "ObjectLogEntryType": ObjectLogEntryType,
    })

    # fed log
    fed_object_log_entries = get_fed_object_log_entries_for_object(
        object_id=object_id
    )
    template_kwargs.update({
        "fed_object_log_entries": fed_object_log_entries,
        "FedObjectLogEntryType": models.FedObjectLogEntryType,
    })

    workflows = get_workflow_references(object, flask_login.current_user.id, actions_by_id)
    template_kwargs.update({
        "workflows": workflows,
        "only_empty_workflows": all(len(workflow) == 0 for workflow in workflows),
    })

    # various getters
    template_kwargs.update({
        "get_object_if_current_user_has_read_permissions": get_object_if_current_user_has_read_permissions,
        "get_fed_object_if_current_user_has_read_permissions": get_fed_object_if_current_user_has_read_permissions,
        "get_user": cache_per_request()(get_user_if_exists),
        "get_location": cache_per_request()(get_location),
        "get_object_location_assignment": get_object_location_assignment,
        "get_project": get_project_if_it_exists,
        "get_action_type": cache_per_request()(get_action_type),
        "get_component": cache_per_request()(get_component),
        "get_shares_for_object": get_shares_for_object,
    })

    if mode in {'', 'inline_edit'} and not flask.current_app.config['DISABLE_INLINE_EDIT'] and user_may_edit:
        template_kwargs.update({
            "errors": {},
            "form_data": {},
            "form": ObjectForm(),
            "mode": 'edit',
        })

        template_kwargs.update(get_object_form_template_kwargs(object_id))

        return flask.render_template(
            'objects/inline_edit/inline_edit_base.html',
            template_mode="inline_edit",
            **template_kwargs
        )
    else:
        return flask.render_template(
            'objects/view/base.html',
            template_mode="view",
            **template_kwargs
        )


@frontend.route('/objects/<int:object_id>/related_objects_entries')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def related_objects_entries(object_id: int) -> FlaskResponseT:
    object = get_object(object_id)
    if object.action_id is None:
        return flask.abort(404)
    action = get_action(object.action_id)
    if not action or not action.type or not action.type.enable_related_objects:
        return flask.abort(404)
    related_objects_subtrees = logic.object_relationships.gather_related_object_subtrees(object_id)
    related_object_refs_by_object_id = {}
    for object_ref in related_objects_subtrees:
        if object_ref.is_local:
            related_object_refs_by_object_id[object_ref.object_id] = object_ref
    related_object_ids = set(related_object_refs_by_object_id)
    related_objects = logic.object_permissions.get_objects_with_permissions(
        user_id=flask_login.current_user.id,
        permissions=Permissions.READ,
        object_ids=tuple(related_object_ids),
        name_only=True
    )
    related_object_by_object_ref = {
        related_object_refs_by_object_id[related_object.object_id]: related_object
        for related_object in related_objects
    }
    related_objects_object_refs = list(related_objects_subtrees)
    related_objects_index_by_object_ref = {
        object_ref: index
        for index, object_ref in enumerate(related_objects_object_refs)
    }

    return flask.render_template(
        'objects/view/related_objects_entries.html',
        related_objects_subtrees=related_objects_subtrees,
        related_object_by_object_ref=related_object_by_object_ref,
        object_ref=logic.object_relationships.ObjectRef(object_id=object_id, component_uuid=None, eln_object_url=None, eln_source_url=None),
        related_objects_object_refs=related_objects_object_refs,
        related_objects_index_by_object_ref=related_objects_index_by_object_ref,
    )


@frontend.route('/objects/<int:object_id>/dc.rdf')
@frontend.route('/objects/<int:object_id>/versions/<int:version_id>/dc.rdf')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_rdf(object_id: int, version_id: typing.Optional[int] = None) -> FlaskResponseT:
    rdf_xml = logic.rdf.generate_rdf(flask_login.current_user.id, object_id, version_id)
    return flask.Response(
        rdf_xml,
        mimetype='application/rdf+xml',
    )


@frontend.route('/objects/<int:object_id>/label')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def print_object_label(object_id: int) -> FlaskResponseT:
    mode = flask.request.args.get('mode', 'mixed')
    action_id = get_object(object_id).action_id

    action = get_action(action_id) if action_id else None
    if not (action and action.type and action.type.enable_labels):
        flask.abort(403)

    only_id_qr_code = False
    add_label_number = False
    add_maximum_label_number = False
    create_only_qr_codes = False
    show_id_on_label = True
    label_quantity = 1
    label_dimension = None
    if mode == 'fixed-width':
        create_mixed_labels = False
        create_long_labels = False
        include_qrcode_in_long_labels = False
        paper_format = flask.request.args.get('width-paper-format', '')
        if paper_format not in PAGE_SIZES:
            paper_format = DEFAULT_PAPER_FORMAT
        maximum_width = math.floor(PAGE_SIZES[paper_format][0] - 2 * HORIZONTAL_LABEL_MARGIN)
        maximum_height = math.floor(PAGE_SIZES[paper_format][1] - 2 * VERTICAL_LABEL_MARGIN)
        ghs_classes_side_by_side = 'side-by-side' in flask.request.args
        label_minimum_width = 20.0
        if ghs_classes_side_by_side:
            label_minimum_width = 40
        try:
            label_width = float(flask.request.args.get('label-width', '20'))
        except ValueError:
            label_width = 0
        if math.isnan(label_width):
            label_width = 0
        if label_width < label_minimum_width:
            label_width = label_minimum_width
        if label_width > maximum_width:
            label_width = maximum_width
        try:
            label_minimum_height = float(flask.request.args.get('label-minimum-height', '0'))
        except ValueError:
            label_minimum_height = 0
        if math.isnan(label_minimum_height):
            label_minimum_height = 0
        if label_minimum_height < 0:
            label_minimum_height = 0
        if label_minimum_height > maximum_height:
            label_minimum_height = maximum_height
        qrcode_width = 18
        centered = 'centered' in flask.request.args
    elif mode == 'minimum-height':
        create_mixed_labels = False
        create_long_labels = True
        paper_format = flask.request.args.get('height-paper-format', '')
        if paper_format not in PAGE_SIZES:
            paper_format = DEFAULT_PAPER_FORMAT
        maximum_width = math.floor(PAGE_SIZES[paper_format][0] - 2 * HORIZONTAL_LABEL_MARGIN)
        include_qrcode_in_long_labels = 'include-qrcode' in flask.request.args
        label_width = 0
        label_minimum_height = 0
        try:
            label_minimum_width = float(flask.request.args.get('label-minimum-width', '0'))
        except ValueError:
            label_minimum_width = 0
        if math.isnan(label_minimum_width):
            label_minimum_width = 0
        if label_minimum_width < 0:
            label_minimum_width = 0
        if label_minimum_width > maximum_width:
            label_minimum_width = maximum_width
        qrcode_width = 0
        ghs_classes_side_by_side = False
        centered = False
    elif mode == 'qr-code-width':
        create_mixed_labels = False
        create_long_labels = False
        create_only_qr_codes = True
        only_id_qr_code = flask.request.args.get('qr-code-content', 'object-url') == "object-id"
        show_id_on_label = 'show-id-on-label' in flask.request.args
        include_qrcode_in_long_labels = False
        add_label_number = 'add-label-number' in flask.request.args
        add_maximum_label_number = 'add-maximum-label-number' in flask.request.args
        paper_format = flask.request.args.get('qr-code-paper-format', '')
        label_dimension_template = flask.request.args.get('qr-code-paper-dimension-template', 'default')

        if paper_format not in PAGE_SIZES:
            paper_format = DEFAULT_PAPER_FORMAT
        try:
            qrcode_width = int(flask.request.args.get('qr-code-width', '7'))
        except ValueError:
            qrcode_width = 7
        if qrcode_width < 4:
            qrcode_width = 4
        elif qrcode_width > 150:
            qrcode_width = 150

        try:
            label_quantity = int(flask.request.args.get('qr-code-number', '1'))
        except ValueError:
            label_quantity = 1
        if label_quantity <= 0:
            label_quantity = 1
        elif label_quantity > 1000:
            label_quantity = 1000

        try:
            label_dimension_idx = int(label_dimension_template)
            label_dimension = flask.current_app.config['LABEL_PAPER_FORMATS'][label_dimension_idx]
            qrcode_width = label_dimension['qr_code_width']
            paper_format = PAGE_SIZE_KEYS[label_dimension['paper_format']]
        except (ValueError, KeyError):
            label_dimension = None

        label_width = 0
        label_minimum_width = qrcode_width
        label_minimum_height = 0
        ghs_classes_side_by_side = False
        centered = False
    else:
        create_mixed_labels = True
        create_long_labels = False
        include_qrcode_in_long_labels = False
        paper_format = flask.request.args.get('mixed-paper-format', '')
        if paper_format not in PAGE_SIZES:
            paper_format = DEFAULT_PAPER_FORMAT
        label_width = 0
        label_minimum_height = 0
        qrcode_width = 0
        label_minimum_width = 0
        ghs_classes_side_by_side = False
        centered = False

    object = get_object(object_id=object_id)
    object_log_entries = object_log.get_object_log_entries(object_id=object_id, user_id=flask_login.current_user.id)
    for object_log_entry in object_log_entries:
        if object_log_entry.type in (ObjectLogEntryType.CREATE_OBJECT, ObjectLogEntryType.CREATE_BATCH):
            creation_date = object_log_entry.utc_datetime.strftime('%Y-%m-%d')
            creation_user = get_user(object_log_entry.user_id).get_name(include_id=False)
            break
    else:
        if object.version_id == 0:
            original_object = object
        else:
            try:
                original_object = get_object(object_id=object_id, version_id=0)
            except errors.ObjectVersionDoesNotExistError:
                original_object = object
        if original_object.utc_datetime is not None:
            creation_date = original_object.utc_datetime.strftime('%Y-%m-%d')
        else:
            creation_date = _('Unknown')
        if original_object.user_id is not None:
            creation_user = get_user(original_object.user_id).get_name(include_id=False)
        else:
            creation_user = _('Unknown')
    if object.data is not None and 'created' in object.data and '_type' in object.data['created'] and object.data['created']['_type'] == 'datetime':
        creation_date = object.data['created']['utc_datetime'].split(' ')[0]
    object_name = get_translated_text(object.name)

    object_url = flask.url_for('.object', object_id=object_id, _external=True)

    if object.data is not None and 'hazards' in object.data and '_type' in object.data['hazards'] and object.data['hazards']['_type'] == 'hazards':
        hazards = object.data['hazards']['hazards']
    else:
        hazards = []

    if 'custom-qr-code-texts-json' in flask.request.args:
        try:
            custom_qr_code_texts = json.loads(flask.request.args['custom-qr-code-texts-json'])
        except ValueError:
            return flask.abort(400)
        if custom_qr_code_texts is not None and (not isinstance(custom_qr_code_texts, dict) or not all(isinstance(value, str) for value in custom_qr_code_texts.values())):
            return flask.abort(400)
    else:
        custom_qr_code_texts = None

    pdf_data = create_labels(
        object_id=object_id,
        object_name=object_name,
        object_url=object_url,
        creation_user=creation_user,
        creation_date=creation_date,
        ghs_classes=hazards,
        paper_format=paper_format,
        create_mixed_labels=create_mixed_labels,
        create_long_labels=create_long_labels,
        create_only_qr_codes=create_only_qr_codes,
        include_qrcode_in_long_labels=include_qrcode_in_long_labels,
        label_width=label_width,
        label_minimum_height=label_minimum_height,
        label_minimum_width=label_minimum_width,
        label_quantity=label_quantity,
        qrcode_width=qrcode_width,
        ghs_classes_side_by_side=ghs_classes_side_by_side,
        centered=centered,
        only_id_qr_code=only_id_qr_code,
        add_label_number=add_label_number,
        add_maximum_label_number=add_maximum_label_number,
        show_id_on_label=show_id_on_label,
        label_dimension=label_dimension,
        custom_qr_code_texts=custom_qr_code_texts
    )
    return flask.send_file(
        io.BytesIO(pdf_data),
        mimetype='application/pdf',
        max_age=-1
    )


@frontend.route("/multiselect_labels", methods=["GET"])
def multiselect_labels() -> FlaskResponseT:
    generate_labels_form = GenerateLabelsForm(flask.request.args)

    if not generate_labels_form.validate():
        flask.abort(400)

    label_variant = generate_labels_form.form_variant.data
    object_ids = list(map(int, generate_labels_form.objects.data.split(',')))
    paper_format = generate_labels_form.paper_size.data

    for object_id in object_ids:
        object = logic.objects.get_object(object_id)
        if object_id is not None and logic.object_permissions.get_user_object_permissions(object_id, user_id=flask_login.current_user.id) >= Permissions.READ:
            if object.action_id:
                action = logic.actions.get_action(object.action_id)
            else:
                action = None
            if action is None or action.type is None or not action.type.enable_labels:
                flask.abort(403)
        else:
            flask.abort(403)

    if paper_format not in PAGE_SIZES:
        paper_format = DEFAULT_PAPER_FORMAT

    maximum_width = math.floor(PAGE_SIZES[paper_format][0] - 2 * HORIZONTAL_LABEL_MARGIN)
    maximum_height = math.floor(PAGE_SIZES[paper_format][1] - 2 * VERTICAL_LABEL_MARGIN)

    qr_code_width = 18.0
    min_label_width = 0.0
    min_label_height = 0.0
    label_width = 0.0

    centered = False
    create_mixed_labels = False
    create_long_labels = False
    include_qrcode_in_long_labels = False
    qr_ghs_side_by_side = False

    try:
        labels_per_object = int(generate_labels_form.labels_per_object.data)
    except ValueError:
        labels_per_object = 1
    if labels_per_object < 1:
        labels_per_object = 1

    if label_variant == 'fixed-width':
        min_label_height = generate_labels_form.min_label_height.data

        qr_ghs_side_by_side = generate_labels_form.qr_ghs_side_by_side.data
        centered = generate_labels_form.center_qr_ghs.data

        label_minimum_width = 40 if qr_ghs_side_by_side else 20
        try:
            label_width = float(generate_labels_form.label_width.data)
        except ValueError:
            label_width = 0

        if math.isnan(label_width):
            label_width = 0
        if label_width < label_minimum_width:
            label_width = label_minimum_width
        if label_width > maximum_width:
            label_width = maximum_width

        try:
            min_label_height = float(min_label_height)
        except ValueError:
            min_label_height = 0
        if math.isnan(min_label_height):
            min_label_height = 0
        if min_label_height < 0:
            min_label_height = 0
        if min_label_height > maximum_height:
            min_label_height = maximum_height

    elif label_variant == 'minimal-height':
        create_long_labels = True
        include_qrcode_in_long_labels = generate_labels_form.include_qr.data

        try:
            min_label_width = float(generate_labels_form.min_label_width.data)
        except ValueError:
            min_label_width = 0
        if math.isnan(min_label_width):
            min_label_width = 0
        if min_label_width < 0:
            min_label_width = 0
        if min_label_width > maximum_width:
            min_label_width = maximum_width

    elif label_variant == 'mixed-formats':
        create_mixed_labels = True

    object_specifications = {}
    for object_id in object_ids:
        object = get_object(object_id=object_id)

        object_log_entries = object_log.get_object_log_entries(object_id=object_id, user_id=flask_login.current_user.id)
        for object_log_entry in object_log_entries:
            if object_log_entry.type in (ObjectLogEntryType.CREATE_OBJECT, ObjectLogEntryType.CREATE_BATCH):
                creation_date = object_log_entry.utc_datetime.strftime("%Y-%m-%d")
                creation_user = get_user(object_log_entry.user_id).get_name(include_id=False)
                break
        else:
            if object.version_id == 0:
                original_object = object
            else:
                try:
                    original_object = get_object(object_id=object_id, version_id=0)
                except errors.ObjectVersionDoesNotExistError:
                    original_object = object
            if original_object.utc_datetime is not None:
                creation_date = original_object.utc_datetime.strftime("%Y-%m-%d")
            else:
                creation_date = _('Unknown')
            if original_object.user_id is not None:
                creation_user = get_user(original_object.user_id).get_name(include_id=False)
            else:
                creation_user = _('Unknown')

        hazards = []
        if object.data is not None:
            if 'created' in object.data and '_type' in object.data['created'] and object.data['created']['_type'] == 'datetime':
                creation_date = object.data['created']['utc_datetime'].split(' ')[0]

            if 'hazards' in object.data and '_type' in object.data['hazards'] and object.data['hazards']['_type'] == 'hazards':
                hazards = object.data['hazards']['hazards']

        object_specifications[object_id] = {
            "object_name": get_translated_text(object.name) if object.name else _('Unknown Object'),
            "object_url": flask.url_for('.object', object_id=object_id, _external=True),
            "creation_user": creation_user,
            "creation_date": creation_date,
            "ghs_classes": hazards
        }

    pdf_data = create_multiple_labels(
        object_specifications=object_specifications,
        quantity=labels_per_object,
        label_width=label_width,
        min_label_width=min_label_width,
        min_label_height=min_label_height,
        qr_code_width=qr_code_width,
        paper_format=paper_format,
        create_mixed_labels=create_mixed_labels,
        create_long_labels=create_long_labels,
        include_qrcode_in_long_labels=include_qrcode_in_long_labels,
        ghs_classes_side_by_side=qr_ghs_side_by_side,
        centered=centered
    )

    return flask.send_file(io.BytesIO(pdf_data), mimetype='application/pdf', max_age=-1)


@frontend.route('/objects/<int:object_id>/comments/', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def post_object_comments(object_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None:
        flask.flash(_('Commenting on imported objects is not yet supported.'), 'error')
        return flask.abort(403)
    if object.action_id is None:
        return flask.abort(403)
    comment_form = CommentForm()
    if comment_form.validate_on_submit():
        content = comment_form.content.data
        comments.create_comment(object_id=object_id, user_id=flask_login.current_user.id, content=content)
        flask.flash(_('Successfully posted a comment.'), 'success')
    else:
        flask.flash(_('Please enter a comment text.'), 'error')
    return flask.redirect(flask.url_for('.object', object_id=object_id))


@frontend.route('/objects/<int:object_id>/locations/', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def post_object_location(object_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None:
        flask.flash(_('Assigning locations to imported objects is not yet supported.'), 'error')
        return flask.abort(403)
    if object.action_id is None:
        return flask.abort(403)
    location_form = ObjectLocationAssignmentForm()
    location_form.location.choices = [('-1', '—')] + [
        (str(location.id), get_location_name(location, include_id=True, has_read_permissions=True))
        for location in get_locations_with_user_permissions(flask_login.current_user.id, Permissions.READ)
        if location.enable_object_assignments and (location.type is None or location.type.enable_object_assignments)
    ]
    possible_responsible_users = [('-1', '—')]
    for user in get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN']):
        possible_responsible_users.append((str(user.id), f'{user.name} (#{user.id})'))
    location_form.responsible_user.choices = possible_responsible_users
    location_id: typing.Optional[int]
    responsible_user_id: typing.Optional[int]
    if location_form.validate_on_submit():
        location_id = int(location_form.location.data)
        if location_id < 0:
            location_id = None
        responsible_user_id = int(location_form.responsible_user.data)
        if responsible_user_id < 0:
            responsible_user_id = None
        description = location_form.description.data
        try:
            description = json.loads(description)
        except Exception:
            description = {}
        valid_description = {'en': ''}
        for language_code, description_text in description.items():
            if not isinstance(language_code, str):
                continue
            try:
                language = get_language_by_lang_code(language_code)
            except errors.LanguageDoesNotExistError:
                continue
            if not language.enabled_for_input:
                continue
            valid_description[language_code] = description_text
        description = valid_description
        if location_id is not None or responsible_user_id is not None:
            try:
                assign_location_to_object(object_id, location_id, responsible_user_id, flask_login.current_user.id, description)
            except errors.ExceedingLocationCapacityError:
                flask.flash(_('The selected location does not have the capacity to store this object.'), 'error')
            else:
                flask.flash(_('Successfully assigned a new location to this object.'), 'success')
        else:
            flask.flash(_('Please select a location or a responsible user.'), 'error')
    else:
        flask.flash(_('Please select a location or a responsible user.'), 'error')
    return flask.redirect(flask.url_for('.object', object_id=object_id))


@frontend.route('/objects/<int:object_id>/publications/', methods=['POST'])
@object_permissions_required(Permissions.WRITE)
def post_object_publication(object_id: int) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    object = get_object(object_id)
    if object.fed_object_id is not None:
        flask.flash(_('Assigning publications to imported objects is not yet supported.'), 'error')
        return flask.abort(403)
    if object.action_id is None:
        return flask.abort(403)
    publication_form = ObjectPublicationForm()
    if publication_form.validate_on_submit():
        doi = publication_form.doi.data
        title = publication_form.title.data
        object_name = publication_form.object_name.data
        if title is not None:
            title = title.strip()
        if not title:
            title = None
        if object_name is not None:
            object_name = object_name.strip()
        if not object_name:
            object_name = None

        existing_publication: typing.Optional[logic.publications.Publication] = None
        for publication in logic.publications.get_publications_for_object(object_id):
            if publication.doi == doi:
                existing_publication = publication
                break
        if existing_publication is not None and existing_publication.title == title and existing_publication.object_name == object_name:
            flask.flash(_('This object has already been linked to this publication.'), 'info')
        else:
            logic.publications.link_publication_to_object(user_id=flask_login.current_user.id, object_id=object_id, doi=doi, title=title, object_name=object_name)
            if existing_publication is None:
                flask.flash(_('Successfully linked this object to a publication.'), 'success')
            else:
                flask.flash(_('Successfully updated the information for this publication.'), 'success')
    else:
        flask.flash(_('Please enter a valid DOI for the publication you want to link this object to.'), 'error')
    return flask.redirect(flask.url_for('.object', object_id=object_id))


@frontend.route('/objects/<int:object_id>/export')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def export_data(object_id: int) -> FlaskResponseT:
    object_ids = [object_id]
    file_extension = flask.request.args.get('format', '.pdf')
    if file_extension != '.pdf' and file_extension not in logic.export.FILE_FORMATS:
        return flask.abort(400)
    if 'object_ids' in flask.request.args:
        try:
            object_ids = json.loads(flask.request.args['object_ids'])
            object_ids = [int(i) for i in object_ids]
            if any((Permissions.READ not in get_user_object_permissions(i, flask_login.current_user.id)) for i in object_ids):
                return flask.abort(400)
        except Exception:
            return flask.abort(400)
        if not object_ids:
            return flask.abort(400)
    if file_extension == '.pdf':
        sections = pdfexport.SECTIONS
        if 'sections' in flask.request.args:
            try:
                sections = sections.intersection(json.loads(flask.request.args['sections']))
            except Exception:
                return flask.abort(400)
        if 'language' in flask.request.args:
            try:
                lang_code = flask.request.args['language']
                if lang_code not in logic.locale.SUPPORTED_LOCALES:
                    raise ValueError()
                language = logic.languages.get_language_by_lang_code(lang_code)
                if not language.enabled_for_user_interface:
                    raise ValueError()
            except Exception:
                lang_code = 'en'
        else:
            lang_code = 'en'
        pdf_data = pdfexport.create_pdfexport(object_ids, sections, lang_code)
        file_bytes = io.BytesIO(pdf_data)
    elif file_extension in logic.export.FILE_FORMATS:
        file_bytes = io.BytesIO(logic.export.FILE_FORMATS[file_extension][1](flask_login.current_user.id, object_ids=object_ids))
    else:
        file_bytes = None
    if file_bytes:
        return flask.Response(
            file_bytes,
            200,
            headers={
                'Content-Disposition': f'attachment; filename=sampledb_export{file_extension}',
                'Content-Type': 'application/pdf' if file_extension == '.pdf' else logic.export.FILE_FORMATS[file_extension][2]
            }
        )
    return flask.abort(500)


@frontend.route('/objects/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_object() -> FlaskResponseT:
    check_current_user_is_not_readonly()

    action_id_str = flask.request.args.get('action_id', None)
    if action_id_str is not None:
        try:
            action_id = int(action_id_str)
        except ValueError:
            action_id = None
    else:
        action_id = None

    previous_object_id_str = flask.request.args.get('previous_object_id', None)
    if previous_object_id_str is not None:
        try:
            previous_object_id = int(previous_object_id_str)
        except ValueError:
            previous_object_id = None
    else:
        previous_object_id = None

    if not action_id and not previous_object_id:
        return flask.abort(404)

    previous_object = None
    if previous_object_id:
        try:
            previous_object = get_object(previous_object_id)
        except errors.ObjectDoesNotExistError:
            return flask.abort(404)
        if Permissions.READ not in get_user_object_permissions(user_id=flask_login.current_user.id, object_id=previous_object_id):
            flask.flash(_("You do not have the required permissions to use this object as a template."), 'error')
            return flask.abort(403)
        if action_id:
            if action_id != previous_object.action_id:
                flask.flash(_("This object was created with a different action."), 'error')
                return flask.abort(400)
        else:
            action_id = previous_object.action_id
    if action_id is None:
        return flask.abort(404)
    else:
        try:
            action = get_action(action_id)
        except errors.ActionDoesNotExistError:
            flask.flash(_("This action does not exist."), 'error')
            return flask.abort(404)
        if Permissions.READ not in get_user_action_permissions(action_id, user_id=flask_login.current_user.id):
            flask.flash(_("You do not have the required permissions to use this action."), 'error')
            return flask.abort(403)
        if action.type_id is None or action.schema is None:
            flask.flash(_("Creating objects with this action has been disabled."), 'error')
            return flask.redirect(flask.url_for('.action', action_id=action_id))

    fields_selected = flask.request.args.get('fields_selected', None) is not None

    placeholder_data: typing.Optional[typing.Dict[typing.Sequence[typing.Union[int, str]], typing.Any]] = {}
    possible_properties: typing.Dict[str, typing.Any] = {}

    passed_object_id_strs = flask.request.args.getlist('object_id')
    if passed_object_id_strs:
        try:
            passed_object_ids = [
                int(passed_object_id_str)
                for passed_object_id_str in passed_object_id_strs
            ]
        except ValueError:
            passed_object_ids = []
        else:
            if any(passed_object_id <= 0 for passed_object_id in passed_object_ids):
                passed_object_ids = []
    else:
        passed_object_ids = []
    if passed_object_ids:
        try:
            passed_objects = [
                get_object(passed_object_id)
                for passed_object_id in passed_object_ids
            ]
        except errors.ObjectDoesNotExistError:
            passed_object_ids = []
            passed_objects = []
    else:
        passed_objects = []
    if passed_object_ids and passed_objects and action is not None and action.schema is not None:
        passed_object_actions = [
            logic.actions.get_action(passed_object.action_id) if passed_object.action_id is not None else None
            for passed_object in passed_objects
        ]
        passed_object_action_types = [
            passed_object_action.type if passed_object_action is not None else None
            for passed_object_action in passed_object_actions
        ]

        allowed_types = ["object_reference"]
        if all(passed_object_action_types):
            if all(passed_object_action_type is not None and models.ActionType.MEASUREMENT in [passed_object_action_type.id, passed_object_action_type.fed_id] for passed_object_action_type in passed_object_action_types):
                allowed_types.append("measurement")

            if all(passed_object_action_type is not None and models.ActionType.SAMPLE_CREATION in [passed_object_action_type.id, passed_object_action_type.fed_id] for passed_object_action_type in passed_object_action_types):
                allowed_types.append("sample")

        def is_property_schema_compatible(
                property_schema: typing.Dict[str, typing.Any],
                allowed_types: typing.Sequence[str],
                passed_object_actions: typing.Sequence[typing.Optional[logic.actions.Action]]
        ) -> bool:
            property_type = property_schema.get('type', '')
            property_action_id = property_schema.get('action_id', None)
            property_action_type_id = property_schema.get('action_type_id', None)

            if property_type not in allowed_types:
                return False

            if property_action_id:
                valid_action_ids = [property_action_id] if type(property_action_id) is int else property_action_id
                if valid_action_ids and any(passed_object_action is None or passed_object_action.id not in valid_action_ids for passed_object_action in passed_object_actions):
                    return False

            if property_action_type_id:
                valid_action_type_ids = [property_action_type_id] if type(property_action_type_id) is int else property_action_type_id
                if valid_action_type_ids and any(passed_object_action is None or passed_object_action.type is None or passed_object_action.type_id not in valid_action_type_ids for passed_object_action in passed_object_actions):
                    return False

            return True

        possible_properties = {}

        for property_name, property_schema in action.schema.get('properties', {}).items():
            if len(passed_object_ids) == 1 and is_property_schema_compatible(property_schema, allowed_types, passed_object_actions):
                possible_properties[property_name] = property_schema
            if property_schema.get('type') == 'array' and is_property_schema_compatible(property_schema['items'], allowed_types, passed_object_actions):
                possible_properties[property_name] = property_schema

        passed_object_property_names: typing.List[str] = []

        if len(possible_properties) == 0:
            flask.flash(_('The selected object cannot be used in this action.')
                        if len(passed_object_ids) == 1
                        else _('The selected objects cannot be used in this action.'), 'error')

        elif len(possible_properties) == 1:
            passed_object_property_names.extend(possible_properties)
        elif len(possible_properties) > 1:
            if fields_selected:
                for property_key in flask.request.form:
                    if property_key in possible_properties:
                        passed_object_property_names.append(property_key)
            else:
                placeholder_data = None
        if placeholder_data is not None:
            for property_key in passed_object_property_names:
                if possible_properties[property_key].get('type') == 'array':
                    placeholder_data[(property_key, )] = [
                        {
                            '_type': possible_properties[property_key].get('type', ''),
                            'object_id': passed_object_id
                        }
                        for passed_object_id in passed_object_ids
                    ]
                    if possible_properties[property_key].get('type') == 'array':
                        num_min_items = possible_properties[property_key].get('minItems', 0)
                        num_passed_objects = len(passed_object_ids)
                        if num_min_items > num_passed_objects:
                            placeholder_data[(property_key,)].extend([None] * (num_min_items - num_passed_objects))
                else:
                    placeholder_data[(property_key, )] = {
                        '_type': possible_properties[property_key].get('type', ''),
                        'object_id': passed_object_ids[0]
                    }

    url_properties: dict[typing.Sequence[str | int], typing.Any] = {}
    prefix = 'properties.'
    for k, [*_vs, v] in flask.request.args.lists():
        if not k.startswith(prefix):
            continue
        path = parse_property_path(k.removeprefix(prefix))
        try:
            schema = iter_schema_path(action.schema, path)
        except Exception:
            flask.flash(_('The path "%(path)s" is not valid for the schema of this action.', path=k), 'error')
            continue
        try:
            url_properties[path] = parse_default_for_subschema(
                schema,
                v,
            )
        except errors.MismatchedUnitError:
            flask.flash(_('The unit of "%(path)s=%(value)s" does not match the unit(s) of the schema: %(units)s', path=k, value=v, units=schema["units"]), 'error')
        except Exception:
            flask.flash(_('The value of "%(path)s=%(value)s" is not valid for this schema: %(schema)s', path=k, value=v, schema=schema), 'error')
    if url_properties:
        if not placeholder_data:
            placeholder_data = {}
        placeholder_data.update(url_properties)

    if previous_object is not None:
        should_upgrade_schema = flask.request.args.get('mode', '') == 'upgrade'
        if action.schema != previous_object.schema and not should_upgrade_schema and flask.current_app.config['DISABLE_OUTDATED_USE_AS_TEMPLATE']:
            return flask.abort(400)
    else:
        should_upgrade_schema = False

    # TODO: check instrument permissions
    return show_object_form(
        object=None,
        action=action,
        previous_object=previous_object,
        should_upgrade_schema=should_upgrade_schema,
        placeholder_data=placeholder_data,
        possible_object_id_properties=possible_properties,
        passed_object_ids=passed_object_ids,
        show_selecting_modal=(not fields_selected)
    )


def parse_property_path(path: str) -> tuple[int | str, ...]:
    def f(v: str) -> str | int:
        try:
            return int(v)
        except Exception:
            return v
    return tuple(f(v) for v in path.split('.'))


def iter_schema_path(schema: dict[str, typing.Any], path: tuple[int | str, ...]) -> dict[str, typing.Any]:
    s = schema
    for k in path:
        while s.get('type', None) in ('object', 'array'):
            if s['type'] == 'object':
                s = s['properties']
            else:
                s = s['items']
        if isinstance(k, int):
            continue
        s = s[k]
    return s


def parse_default_for_subschema(schema: dict[str, typing.Any], value: str) -> dict[str, typing.Any]:
    def quantity() -> dict[str, typing.Any]:
        m = re.match(r'^(\d+\.?\d*)(.*)', value)
        if not m:
            raise ValueError()
        magnitude = m.group(1)
        unit = m.group(2) or '1'
        if unit not in schema['units']:
            raise errors.MismatchedUnitError()
        return {
            'magnitude': float(magnitude),
            'units': unit,
        }

    def simple(
            k: str,
            f: typing.Callable[[str], typing.Any] | None = None,
    ) -> typing.Callable[[], dict[str, typing.Any]]:
        return lambda: {
            k: f(value) if f else value
        }

    functions = {
        'quantity': quantity,
        'text': simple('text'),
        'bool': simple('value', lambda v: v == 'true'),
        'datetime': simple('utc_datetime'),
        'tags': simple('tags', lambda v: v.split(',')),
        'user': simple('user_id', int),
        'object_reference': simple('object_id', int),
        'sample': simple('object_id', int),
        'measurement': simple('object_id', int),
    }

    instance = {
        '_type': schema['type'],
    } | functions[schema['type']]()

    logic.schemas.validate(instance, schema)

    return instance


@frontend.route('/objects/<int:object_id>/timeseries_data/<timeseries_id>')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def download_timeseries_data(object_id: int, timeseries_id: str) -> FlaskResponseT:
    object = get_object(object_id)
    if object.data is None:
        return flask.abort(404)
    if not timeseries_id.startswith('object__') and not timeseries_id.startswith(f'object{object_id}__'):
        return flask.abort(404)
    id_path = timeseries_id.split('__')[1:]
    data_and_schema = get_data_and_schema_by_id_path(object.data, object.schema, typing.cast(typing.List[typing.Union[str, int]], id_path), convert_id_path_elements=True)
    if data_and_schema is None:
        return flask.abort(404)
    data, schema = data_and_schema
    if not isinstance(schema, dict) or schema['type'] != 'timeseries' or not isinstance(data, dict):
        return flask.abort(404)
    csv_io = io.StringIO()
    writer = csv.writer(csv_io, quoting=csv.QUOTE_NONNUMERIC)
    if isinstance(data['data'][0][0], str):
        time_row_name = 'utc_datetime'
    else:
        time_row_name = 'time in s'
    writer.writerow([time_row_name, 'magnitude in ' + str(data['units']), 'magnitude in base units'])
    writer.writerows(data['data'])
    binary_csv_io = io.BytesIO(csv_io.getvalue().encode('utf-8'))
    return flask.send_file(
        binary_csv_io,
        as_attachment=True,
        download_name='timeseries.csv',
        mimetype='text/csv',
        last_modified=object.utc_datetime
    )



# coding: utf-8
"""

"""
import typing
import json

import flask
import flask_login
import werkzeug.datastructures
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, SelectField, SelectMultipleField, BooleanField, IntegerField, FormField, FieldList
from wtforms.validators import DataRequired

from . import frontend
from .permission_forms import set_up_permissions_forms, handle_permission_forms
from .. import logic
from ..logic import errors
from ..logic.components import get_component
from ..logic.locations import Location, create_location, get_location, get_locations_tree, update_location, get_object_location_assignment, confirm_object_responsibility, decline_object_responsibility, get_object_location_assignments, get_location_capacities, get_assigned_object_count_by_action_types
from ..logic.location_log import get_log_entries_for_location
from ..logic.languages import Language, get_language, get_languages, get_language_by_lang_code
from ..logic.security_tokens import verify_token
from ..logic.notifications import mark_notification_for_being_assigned_as_responsible_user_as_read, create_notification_for_a_declined_responsibility_assignment
from ..logic.location_permissions import get_user_location_permissions, get_location_permissions_for_all_users, get_location_permissions_for_users, get_location_permissions_for_groups, get_location_permissions_for_projects, set_location_permissions_for_all_users, set_user_location_permissions
from ..logic.users import get_user, get_users
from ..logic.groups import get_group
from ..logic.projects import get_project
from ..logic.settings import get_user_settings, set_user_settings
from ..logic.topics import get_topics, set_location_topics
from .utils import check_current_user_is_not_readonly, get_location_name, get_groups_form_data, parse_filter_id_params, build_modified_url
from ..utils import FlaskResponseT
from ..logic.utils import get_translated_text
from ..models import Permissions, LocationLogEntryType


LOCATION_LIST_FILTER_PARAMETERS = (
    'location_list_filters',
    'topic_ids'
)


class LocationCapacityForm(FlaskForm):
    action_type_id = SelectField(validators=[DataRequired()], coerce=int, choices=[], validate_choice=False)
    capacity = IntegerField()


class LocationForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    parent_location = SelectField()
    is_public = BooleanField(default=True)
    type = SelectField(validators=[DataRequired()])
    responsible_users = SelectMultipleField()
    is_hidden = BooleanField(default=False)
    enable_object_assignments = BooleanField(default=True)
    capacities = FieldList(FormField(LocationCapacityForm))
    topics = SelectMultipleField()


@frontend.route('/locations/')
@flask_login.login_required
def locations() -> FlaskResponseT:
    locations_map, locations_tree = get_locations_tree()
    permissions_by_id = {
        location_id: get_user_location_permissions(location_id, flask_login.current_user.id)
        for location_id in locations_map
    }

    topics = get_topics()
    topics_by_id = {
        topic.id: topic
        for topic in topics
    }
    valid_topic_ids = [topic.id for topic in topics]

    if 'location_list_filters' in flask.request.args or any(any(flask.request.args.getlist(param)) for param in LOCATION_LIST_FILTER_PARAMETERS):
        (
            success,
            filter_topic_ids,
        ) = _parse_location_list_filters(
            params=flask.request.args,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
    else:
        filter_topic_ids = get_user_settings(user_id=flask_login.current_user.id)['DEFAULT_LOCATION_LIST_FILTERS'].get('filter_topic_ids', [])

    if filter_topic_ids is None or flask.current_app.config['DISABLE_TOPICS']:
        filter_topic_ids = []

    if filter_topic_ids:
        def locations_filter(location: Location) -> bool:
            return location_has_topic_ids(location, frozenset(filter_topic_ids))
        locations_tree = filter_location_tree_by_predicate(locations_tree, locations_map, locations_filter)
    else:
        def locations_filter(location: Location) -> bool:
            return True

    filter_topic_infos = []
    if filter_topic_ids:
        for topic_id in filter_topic_ids:
            topic = topics_by_id[topic_id]
            filter_topic_infos.append({
                'name': get_translated_text(topic.name, default=_('Unnamed Topic')),
                'url': flask.url_for('.topic', topic_id=topic_id)
            })

    filtered_locations_tree = filter_location_tree_by_predicate(locations_tree, locations_map, lambda location: user_has_read_permissions_for_location(location, permissions_by_id))
    has_hidden_locations = filtered_locations_tree != locations_tree
    return flask.render_template(
        'locations/locations.html',
        locations_map=locations_map,
        locations_tree=filtered_locations_tree,
        sort_location_ids_by_name=sort_location_ids_by_name,
        has_hidden_locations=has_hidden_locations,
        permissions_by_id=permissions_by_id,
        Permissions=Permissions,
        get_component=get_component,
        get_user=get_user,
        topics=topics,
        filter_topic_infos=filter_topic_infos,
        filter_topic_ids=filter_topic_ids,
        locations_filter=locations_filter
    )


_LocationTree = typing.Dict[int, '_LocationTree']


def location_has_topic_ids(location: Location, topic_ids: typing.FrozenSet[int]) -> bool:
    location_topic_ids = {
        topic.id
        for topic in location.topics
    }
    return len(location_topic_ids.intersection(topic_ids)) > 0


def user_has_read_permissions_for_location(location: Location, permissions_by_id: typing.Dict[int, Permissions]) -> bool:
    return Permissions.READ in permissions_by_id[location.id]


def filter_location_tree_by_predicate(
        location_tree: _LocationTree,
        locations_map: typing.Dict[int, Location],
        predicate: typing.Callable[[Location], bool]
) -> _LocationTree:
    filtered_locations_tree: _LocationTree = {}
    for location_id in location_tree:
        sub_location_tree = location_tree[location_id]
        if sub_location_tree:
            sub_location_tree = filter_location_tree_by_predicate(sub_location_tree, locations_map, predicate)
        if predicate(locations_map[location_id]) or sub_location_tree:
            filtered_locations_tree[location_id] = sub_location_tree
    return filtered_locations_tree


def _parse_location_list_filters(
        params: werkzeug.datastructures.MultiDict[str, str],
        valid_topic_ids: typing.List[int]
) -> typing.Tuple[
    bool,
    typing.Optional[typing.List[int]],
]:
    FALLBACK_RESULT = False, None
    success, filter_topic_ids = parse_filter_id_params(
        params=params,
        param_aliases=['topic_ids'],
        valid_ids=valid_topic_ids,
        id_map={},
        multi_params_error='',
        parse_error=_('Unable to parse topic IDs.'),
        invalid_id_error=_('Invalid topic ID.')
    )
    if not success:
        return FALLBACK_RESULT

    return (
        True,
        filter_topic_ids,
    )


@frontend.route('/locations/', methods=['POST'])
@flask_login.login_required
def save_location_list_defaults() -> FlaskResponseT:
    if 'save_default_location_filters' in flask.request.form:
        topics = get_topics()
        valid_topic_ids = [topic.id for topic in topics]
        success, filter_topic_ids = _parse_location_list_filters(
            params=flask.request.form,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_LOCATION_LIST_FILTERS': {
                    'filter_topic_ids': filter_topic_ids,
                }
            }
        )
        return flask.redirect(build_modified_url('.locations', blocked_parameters=LOCATION_LIST_FILTER_PARAMETERS))
    if 'clear_default_location_filters' in flask.request.form:
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_LOCATION_LIST_FILTERS': {}
            }
        )
        return flask.redirect(build_modified_url('.locations', blocked_parameters=LOCATION_LIST_FILTER_PARAMETERS))
    return flask.abort(400)


@frontend.route('/locations/<int:location_id>', methods=['GET', 'POST'])
@flask_login.login_required
def location(location_id: int) -> FlaskResponseT:
    try:
        location = get_location(location_id)
    except errors.LocationDoesNotExistError:
        return flask.abort(404)
    permissions = get_user_location_permissions(location_id, flask_login.current_user.id)
    mode = flask.request.args.get('mode', None)
    if mode == 'edit':
        if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS'] and not flask_login.current_user.is_admin:
            flask.flash(_('Only administrators can edit locations.'), 'error')
            return flask.abort(403)
        if location.fed_id is not None:
            flask.flash(_('Editing imported locations is not yet supported.'), 'error')
            return flask.abort(403)
        check_current_user_is_not_readonly()
        if Permissions.WRITE not in permissions:
            flask.flash(_('You do not have permissions to edit this location.'), 'error')
            return flask.abort(403)
        return _show_location_form(location, None, Permissions.GRANT in permissions)
    if Permissions.READ not in permissions:
        flask.flash(_('You do not have permissions to view this location.'), 'error')
        return flask.abort(403)
    locations_map, locations_tree = get_locations_tree()
    ancestors: typing.List[Location] = []
    parent_location = location
    while parent_location.parent_location_id is not None:
        parent_location = get_location(parent_location.parent_location_id)
        ancestors.insert(0, parent_location)
    for ancestor in ancestors:
        locations_tree = locations_tree[ancestor.id]
    locations_tree = locations_tree[location_id]
    permissions_by_id = {
        location_id: get_user_location_permissions(location_id, flask_login.current_user.id)
        for location_id in locations_map
    }
    descendent_location_ids = logic.locations.get_descendent_location_ids(locations_tree)
    if location.type.enable_capacities:
        capacities = get_location_capacities(location.id)
        stored_objects = get_assigned_object_count_by_action_types(location.id)
        action_types = logic.action_types.get_action_types(filter_fed_defaults=True)
    else:
        capacities = None
        stored_objects = None
        action_types = None
    return flask.render_template(
        'locations/location.html',
        locations_map=locations_map,
        locations_tree=locations_tree,
        descendent_location_ids=descendent_location_ids,
        location=location,
        ancestors=ancestors,
        sort_location_ids_by_name=sort_location_ids_by_name,
        permissions=permissions,
        permissions_by_id=permissions_by_id,
        Permissions=Permissions,
        location_log_entries=get_log_entries_for_location(location.id, flask_login.current_user.id) if location.type.show_location_log else None,
        LocationLogEntryType=LocationLogEntryType,
        capacities=capacities,
        stored_objects=stored_objects,
        action_types=action_types,
        get_component=get_component,
        get_user=get_user
    )


@frontend.route('/locations/<int:location_id>/permissions', methods=['GET', 'POST'])
@flask_login.login_required
def location_permissions(location_id: int) -> FlaskResponseT:
    try:
        location = get_location(location_id)
    except errors.LocationDoesNotExistError:
        return flask.abort(404)
    permissions = get_user_location_permissions(location_id, flask_login.current_user.id)
    if Permissions.READ not in permissions:
        return flask.abort(403)
    user_may_edit = Permissions.GRANT in permissions

    all_user_permissions = get_location_permissions_for_all_users(location_id)
    user_permissions = get_location_permissions_for_users(location_id)
    group_permissions = get_location_permissions_for_groups(location_id)
    project_permissions = get_location_permissions_for_projects(location_id)
    if user_may_edit:
        (
            add_user_permissions_form,
            add_group_permissions_form,
            add_project_permissions_form,
            permissions_form
        ) = set_up_permissions_forms(
            resource_permissions=logic.location_permissions.location_permissions,
            resource_id=location_id,
            existing_all_user_permissions=all_user_permissions,
            existing_anonymous_user_permissions=None,
            existing_user_permissions=user_permissions,
            existing_group_permissions=group_permissions,
            existing_project_permissions=project_permissions
        )
        permissions_form.all_user_permissions.choices = [
            (p.name.lower(), p)
            for p in [
                Permissions.NONE, Permissions.READ, Permissions.WRITE
            ]
        ]

        users = logic.users.get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'], exclude_fed=True, exclude_eln_import=True)
        users = [user for user in users if user.id not in user_permissions]
        users.sort(key=lambda user: user.id)

        show_groups_form, groups_treepicker_info = get_groups_form_data(
            basic_group_filter=lambda group: group.id not in group_permissions
        )

        show_projects_form, projects_treepicker_info = get_groups_form_data(
            project_group_filter=lambda group: group.id not in project_permissions
        )

        if flask.request.method.lower() == 'post':
            if handle_permission_forms(
                logic.location_permissions.location_permissions,
                location_id,
                add_user_permissions_form,
                add_group_permissions_form,
                add_project_permissions_form,
                permissions_form,
                user_permissions,
                group_permissions,
                project_permissions
            ):
                flask.flash(_('Successfully updated location permissions.'), 'success')
            else:
                flask.flash(_('Failed to update location permissions.'), 'error')
            return flask.redirect(flask.url_for('.location_permissions', location_id=location_id))
    else:
        permissions_form = None
        users = None
        add_user_permissions_form = None
        add_group_permissions_form = None
        add_project_permissions_form = None
        show_groups_form = False
        groups_treepicker_info = None
        show_projects_form = False
        projects_treepicker_info = None
        if flask.request.method.lower() == 'post':
            flask.flash(_('You need GRANT permissions to edit the permissions for this location.'), 'error')
            return flask.redirect(flask.url_for('.location_permissions', location_id=location_id))

    return flask.render_template(
        'locations/location_permissions.html',
        user_may_edit=user_may_edit,
        location=location,
        Permissions=Permissions,
        project_permissions=project_permissions,
        group_permissions=group_permissions,
        user_permissions=user_permissions,
        all_user_permissions=all_user_permissions,
        form=permissions_form,
        users=users,
        add_user_permissions_form=add_user_permissions_form,
        add_group_permissions_form=add_group_permissions_form,
        add_project_permissions_form=add_project_permissions_form,
        show_groups_form=show_groups_form,
        groups_treepicker_info=groups_treepicker_info,
        show_projects_form=show_projects_form,
        projects_treepicker_info=projects_treepicker_info,
        get_user=get_user,
        get_group=get_group,
        get_project=get_project,
    )


@frontend.route('/locations/new/', methods=['GET', 'POST'])
@flask_login.login_required
def new_location() -> FlaskResponseT:
    if flask.current_app.config['ONLY_ADMINS_CAN_MANAGE_LOCATIONS'] and not flask_login.current_user.is_admin:
        flask.flash(_('Only administrators can create locations.'), 'error')
        return flask.abort(403)
    check_current_user_is_not_readonly()
    parent_location = None
    parent_location_id_str = flask.request.args.get('parent_location_id', None)
    if parent_location_id_str is not None:
        try:
            parent_location_id = int(parent_location_id_str)
        except ValueError:
            parent_location_id = None
    else:
        parent_location_id = None
    if parent_location_id:
        try:
            parent_location = get_location(parent_location_id)
        except errors.LocationDoesNotExistError:
            flask.flash(_('The requested parent location does not exist.'), 'error')
        if parent_location is not None:
            parent_location_permissions = get_user_location_permissions(parent_location_id, flask_login.current_user.id)
            if Permissions.WRITE not in parent_location_permissions:
                flask.flash(_('You do not have the required permissions for the requested parent location.'), 'error')
                parent_location = None
    return _show_location_form(None, parent_location, True)


def _handle_object_location_assignment(
        token_salt: str,
        missing_token_text: str,
        invalid_token_text: str,
        success_text: str,
        callback: typing.Callable[[int], None]
) -> FlaskResponseT:
    token = flask.request.args.get('t', None)
    if token is None:
        flask.flash(missing_token_text, 'error')
        return flask.redirect(flask.url_for('.index'))
    object_location_assignment_id = verify_token(
        token=token,
        salt=token_salt,
        secret_key=flask.current_app.config['SECRET_KEY'],
        expiration=None
    )
    if object_location_assignment_id is None:
        flask.flash(invalid_token_text, 'error')
        return flask.redirect(flask.url_for('.index'))
    try:
        object_location_assignment = get_object_location_assignment(object_location_assignment_id)
    except errors.ObjectLocationAssignmentDoesNotExistError:
        flask.flash(_('This responsibility assignment does not exist.'), 'error')
        return flask.redirect(flask.url_for('.index'))
    if object_location_assignment.responsible_user_id != flask_login.current_user.id:
        flask.flash(_('This responsibility assignment belongs to another user.'), 'error')
        return flask.redirect(flask.url_for('.index'))
    if object_location_assignment.confirmed:
        flask.flash(_('This responsibility assignment has already been confirmed.'), 'success')
    elif object_location_assignment.declined:
        flask.flash(_('This responsibility assignment has already been declined.'), 'success')
    else:
        callback(object_location_assignment_id)
        flask.flash(success_text, 'success')
        mark_notification_for_being_assigned_as_responsible_user_as_read(
            user_id=flask_login.current_user.id,
            object_location_assignment_id=object_location_assignment_id
        )
    return flask.redirect(flask.url_for('.object', object_id=object_location_assignment.object_id))


@frontend.route('/locations/confirm_responsibility')
@flask_login.login_required
def accept_responsibility_for_object() -> FlaskResponseT:
    return _handle_object_location_assignment(
        token_salt='confirm_responsibility',
        missing_token_text=_('The confirmation token is missing.'),
        invalid_token_text=_('The confirmation token is invalid.'),
        success_text=_('You have successfully confirmed this responsibility assignment.'),
        callback=confirm_object_responsibility
    )


@frontend.route('/locations/decline_responsibility')
@flask_login.login_required
def decline_responsibility_for_object() -> FlaskResponseT:
    def _callback(object_location_assignment_id: int) -> None:
        decline_object_responsibility(object_location_assignment_id)
        object_location_assignment = get_object_location_assignment(object_location_assignment_id)
        if object_location_assignment.user_id != object_location_assignment.responsible_user_id and object_location_assignment.user_id is not None:
            # notify the assigning user that the assignment was declined
            create_notification_for_a_declined_responsibility_assignment(object_location_assignment.user_id, object_location_assignment_id)
        all_object_location_assignments = get_object_location_assignments(object_id=object_location_assignment.object_id)
        all_object_location_assignments.sort(key=lambda assignment: (assignment.utc_datetime, assignment.id), reverse=True)
        for assignment in all_object_location_assignments:
            if assignment.confirmed and assignment.responsible_user_id is not None:
                current_responsible_user_id = assignment.responsible_user_id
                break
        else:
            current_responsible_user_id = None
        if current_responsible_user_id not in (None, object_location_assignment.responsible_user_id, object_location_assignment.user_id):
            assert current_responsible_user_id is not None  # needed by mypy but guaranteed by the surrounding if
            # notify the currently responsible user that the assignment was declined
            create_notification_for_a_declined_responsibility_assignment(current_responsible_user_id, object_location_assignment_id)
    return _handle_object_location_assignment(
        token_salt='decline_responsibility',
        missing_token_text=_('The declination token is missing.'),
        invalid_token_text=_('The declination token is invalid.'),
        success_text=_('You have successfully declined this responsibility assignment.'),
        callback=_callback
    )


def sort_location_ids_by_name(location_ids: typing.Iterable[int], location_map: typing.Dict[int, Location]) -> typing.List[int]:
    location_ids = list(location_ids)
    location_ids.sort(key=lambda location_id: get_location_name(location_map[location_id]))
    return location_ids


def _show_location_form(
        location: typing.Optional[Location],
        parent_location: typing.Optional[Location],
        has_grant_permissions: bool
) -> FlaskResponseT:
    english = get_language(Language.ENGLISH)
    name_language_ids: typing.List[int] = []
    description_language_ids: typing.List[int] = []

    class LanguageTranslation(typing.TypedDict):
        language_id: int
        lang_name: str
        name: str
        description: str

    location_translations: typing.List[LanguageTranslation] = []
    if location is not None:
        submit_text = "Save"
        may_change_hidden = flask_login.current_user.is_admin
        may_change_enable_object_assignments = flask_login.current_user.is_admin and (location.type is None or location.type.enable_object_assignments)
    else:
        submit_text = "Create"
        may_change_hidden = False
        may_change_enable_object_assignments = False

    location_types = logic.locations.get_location_types()

    locations_map, locations_tree = get_locations_tree()
    invalid_location_ids = []
    if location is not None:
        invalid_location_ids.append(location.id)
        ancestor_ids: typing.List[int] = []
        _parent_location = location
        while _parent_location.parent_location_id is not None:
            _parent_location = get_location(_parent_location.parent_location_id)
            ancestor_ids.insert(0, _parent_location.id)
        locations_subtree = locations_tree
        for ancestor_id in ancestor_ids:
            locations_subtree = locations_subtree[ancestor_id]
        locations_subtree = locations_subtree[location.id]
        unhandled_descendent_ids_and_subtrees = [(descendent_id, locations_subtree) for descendent_id in locations_subtree]
        while unhandled_descendent_ids_and_subtrees:
            descendent_id, locations_subtree = unhandled_descendent_ids_and_subtrees.pop(0)
            invalid_location_ids.append(descendent_id)
            locations_subtree = locations_subtree[descendent_id]
            for descendent_id in locations_subtree:
                unhandled_descendent_ids_and_subtrees.append((descendent_id, locations_subtree))

    for location_id in locations_map:
        if location_id not in invalid_location_ids:
            if Permissions.WRITE not in get_user_location_permissions(location_id, flask_login.current_user.id):
                invalid_location_ids.append(location_id)

    location_form = LocationForm()
    location_form.parent_location.choices = [('-1', '-')] + [
        (str(location_id), locations_map[location_id].name)
        for location_id in locations_map
        if location_id not in invalid_location_ids and locations_map[location_id].type.enable_sub_locations
    ]
    location_form.topics.choices = [
        (str(topic.id), topic)
        for topic in get_topics()
    ]
    topic_ids_str = flask.request.args.get('topic_ids')
    if topic_ids_str is not None:
        valid_topic_id_strs = [
            topic_id_str
            for topic_id_str, topic in location_form.topics.choices
        ]
        location_form.topics.default = [
            topic_id_str
            for topic_id_str in topic_ids_str.split(',')
            if topic_id_str in valid_topic_id_strs
        ]
    action_types = logic.action_types.get_action_types(filter_fed_defaults=True)
    valid_action_types = [
        action_type
        for action_type in action_types
        if action_type.enable_locations and not action_type.disable_create_objects
    ]
    if has_grant_permissions:
        users = [
            user
            for user in get_users()
            if (user.fed_id is None and not user.is_hidden) or (location is not None and user in location.responsible_users)
        ]
        location_form.responsible_users.choices = [
            (str(user.id), user.get_name())
            for user in users
        ]
    else:
        location_form.responsible_users.choices = []
        users = []
    # filter permitted location types
    location_types = [
        location_type
        for location_type in location_types
        if not location_type.admin_only or flask_login.current_user.is_admin or (location is not None and location.type_id == location_type.id)
    ]
    location_form.type.choices = [
        (str(location_type.id), location_type.name)
        for location_type in location_types
    ]
    location_type_is_fed = {
        str(location_type.id): location_type.fed_id is not None
        for location_type in location_types
    }
    location_type_enable_parent_location = {
        str(location_type.id): location_type.enable_parent_location
        for location_type in location_types
    }
    location_type_enable_responsible_users = {
        str(location_type.id): location_type.enable_responsible_users
        for location_type in location_types
    }
    location_type_enable_capacities = {
        str(location_type.id): location_type.enable_capacities
        for location_type in location_types
    }
    location_is_fed = {
        str(loc.id): loc.fed_id is not None
        for loc in locations_map.values()
        if loc.id not in invalid_location_ids
    }
    previous_parent_location_is_invalid = False
    if not location_form.is_submitted():
        if location is not None and location.parent_location_id:
            location_form.parent_location.data = str(location.parent_location_id)
        elif parent_location is not None:
            location_form.parent_location.data = str(parent_location.id)
        else:
            location_form.parent_location.data = str(-1)
        if location_form.parent_location.data not in [
            value
            for value, label in location_form.parent_location.choices
        ]:
            previous_parent_location_is_invalid = True
            location_form.parent_location.data = str(-1)
        if location is not None:
            location_form.topics.data = [str(topic.id) for topic in location.topics]
        if location is not None:
            location_form.type.data = str(location.type_id)
        else:
            # default to the generic location type
            location_form.type.data = str(logic.locations.LocationType.LOCATION)
        if location is not None and has_grant_permissions:
            location_form.responsible_users.data = [str(user.id) for user in location.responsible_users]
        else:
            location_form.responsible_users.data = []
        if location is not None:
            location_form.is_hidden.data = location.is_hidden
        else:
            location_form.is_hidden.data = False
        if location is not None:
            location_form.enable_object_assignments.data = location.enable_object_assignments
        else:
            location_form.enable_object_assignments.data = True

        if location is not None:
            capacities = get_location_capacities(location.id)
            for action_type_id, capacity in capacities.items():
                location_form.capacities.append_entry({
                    'action_type_id': action_type_id,
                    'capacity': capacity
                })

    form_is_valid = False
    if location_form.validate_on_submit():
        form_is_valid = True

    for action_type in valid_action_types:
        for entry in location_form.capacities.entries:
            if entry.action_type_id.data == action_type.id:
                break
        else:
            location_form.capacities.append_entry({
                'action_type_id': action_type.id,
                'capacity': 0
            })

    if location is not None:
        if location.name is not None:
            for language_code, name in location.name.items():
                language = get_language_by_lang_code(language_code)
                if not language.enabled_for_input:
                    continue
                name_language_ids.append(language.id)
                location_translations.append({
                    'language_id': language.id,
                    'lang_name': get_translated_text(language.names),
                    'name': name,
                    'description': ''
                })

        if location.description is not None:
            for language_code, description in location.description.items():
                language = get_language_by_lang_code(language_code)
                if not language.enabled_for_input:
                    continue
                description_language_ids.append(language.id)
                for translation in location_translations:
                    if language.id == translation['language_id']:
                        translation['description'] = description
                        break
                else:
                    location_translations.append({
                        'language_id': language.id,
                        'lang_name': get_translated_text(language.names),
                        'name': '',
                        'description': description
                    })

    if form_is_valid:
        try:
            translations = json.loads(location_form.translations.data)
            names = {}
            descriptions = {}
            if not translations:
                raise ValueError(_('Please enter at least an english name.'))
            for translation in translations:
                name = translation['name'].strip()
                description = translation['description'].strip()
                language_id = int(translation['language_id'])
                if language_id == Language.ENGLISH:
                    if name == '':
                        raise ValueError(_('Please enter at least an english name.'))
                elif name == '' and description == '':
                    continue
                language = get_language(language_id)
                if language.enabled_for_input:
                    names[language.lang_code] = name
                    descriptions[language.lang_code] = description
                else:
                    location_form.translations.errors.append(_('One of these languages is not supported.'))
        except errors.LanguageDoesNotExistError:
            location_form.translations.errors.append(_('One of these languages is not supported.'))
        except Exception as e:
            location_form.translations.errors.append(str(e))
        form_is_valid = not location_form.translations.errors
    if form_is_valid:
        try:
            parent_location_id = int(location_form.parent_location.data)
        except ValueError:
            parent_location_id = None
        if parent_location_id == -1:
            parent_location_id = None
        if parent_location_id is not None:
            try:
                parent_location = logic.locations.get_location(parent_location_id)
            except errors.LocationDoesNotExistError:
                parent_location_id = None
                parent_location = None
                form_is_valid = False
        else:
            parent_location = None
    if form_is_valid:
        try:
            location_type_id = int(location_form.type.data)
        except ValueError:
            location_type_id = None
        if location_type_id is not None:
            try:
                location_type = logic.locations.get_location_type(location_type_id)
            except errors.LocationTypeDoesNotExistError:
                location_type_id = None
                location_type = None
                form_is_valid = False
        else:
            location_type = None
            form_is_valid = False
            location_form.type.errors.append(_('Please select a valid location type.'))
    else:
        location_type = None
        location_type_id = None
    if location_type is None or location_type_id is None:
        form_is_valid = False
    responsible_user_ids = []
    if form_is_valid and has_grant_permissions:
        valid_user_ids = [
            user.id
            for user in users
        ]
        for user_id in location_form.responsible_users.data:
            try:
                user_id = int(user_id)
            except ValueError:
                continue
            if user_id not in responsible_user_ids and user_id in valid_user_ids:
                responsible_user_ids.append(user_id)
    if form_is_valid:
        assert location_type is not None  # mypy does not infer this, but it is guaranteed by the code above
        if location_type.admin_only and not flask_login.current_user.is_admin and location is None:
            location_form.type.errors.append(_('Only administrators may create locations of this type.'))
            form_is_valid = False
        if parent_location is not None:
            if not location_type.enable_parent_location:
                location_form.parent_location.errors.append(_('Locations of this type may not have a parent location.'))
                form_is_valid = False
            if not parent_location.type.enable_sub_locations:
                location_form.parent_location.errors.append(_('This location may not have sub locations.'))
                form_is_valid = False
        if not location_type.enable_responsible_users and responsible_user_ids:
            location_form.responsible_users.errors.append(_('This location may not have responsible users.'))
            form_is_valid = False
    if form_is_valid:
        assert location_type is not None  # mypy does not infer this, but it is guaranteed by the code above
        if location_type.enable_capacities:
            valid_action_type_ids = {
                action_type.id
                for action_type in valid_action_types
            }
            handled_action_types = set()
            for entry in location_form.capacities.entries:
                if entry.action_type_id.data not in valid_action_type_ids:
                    form_is_valid = False
                elif entry.action_type_id.data in handled_action_types:
                    form_is_valid = False
                else:
                    handled_action_types.add(entry.action_type_id.data)
                if entry.capacity.data is not None and not 0 <= entry.capacity.data <= 1000000000:
                    form_is_valid = False
    if form_is_valid:
        assert location_type is not None  # mypy does not infer this, but it is guaranteed by the code above
        assert location_type_id is not None  # mypy does not infer this, but it is guaranteed by the code above
        if location is None:
            location = create_location(
                name=names,
                description=descriptions,
                parent_location_id=parent_location_id,
                user_id=flask_login.current_user.id,
                type_id=location_type_id
            )
            if location_form.is_public.data:
                set_location_permissions_for_all_users(location.id, Permissions.WRITE)
            else:
                set_user_location_permissions(location.id, flask_login.current_user.id, Permissions.GRANT)
            flask.flash(_('The location was created successfully.'), 'success')
        else:
            update_location(
                location_id=location.id,
                name=names,
                description=descriptions,
                parent_location_id=parent_location_id,
                user_id=flask_login.current_user.id,
                type_id=location_type_id,
                is_hidden=location_form.is_hidden.data if may_change_hidden else location.is_hidden,
                enable_object_assignments=location_form.enable_object_assignments.data if may_change_enable_object_assignments else location.enable_object_assignments,
            )
            flask.flash(_('The location was updated successfully.'), 'success')
        if has_grant_permissions:
            logic.locations.set_location_responsible_users(location.id, responsible_user_ids)
        if not flask.current_app.config['DISABLE_TOPICS']:
            topic_ids = [
                int(topic_id)
                for topic_id in location_form.topics.data
            ]
            set_location_topics(location.id, topic_ids)
        if location_type.enable_capacities:
            for entry in location_form.capacities.entries:
                logic.locations.set_location_capacity(location.id, entry.action_type_id.data, entry.capacity.data)
        else:
            logic.locations.clear_location_capacities(location.id)
        return flask.redirect(flask.url_for('.location', location_id=location.id))
    if english.id not in description_language_ids:
        description_language_ids.append(english.id)
    return flask.render_template(
        'locations/location_form.html',
        location=location,
        location_types=location_types,
        location_form=location_form,
        submit_text=submit_text,
        ENGLISH=english,
        languages=get_languages(only_enabled_for_input=True),
        translations=location_translations,
        name_language_ids=name_language_ids,
        description_language_ids=description_language_ids,
        location_is_fed=location_is_fed,
        location_type_is_fed=location_type_is_fed,
        location_type_enable_parent_location=location_type_enable_parent_location,
        location_type_enable_responsible_users=location_type_enable_responsible_users,
        location_type_enable_capacities=location_type_enable_capacities,
        valid_action_types=valid_action_types,
        previous_parent_location_is_invalid=previous_parent_location_is_invalid,
        has_grant_permissions=has_grant_permissions,
        may_change_hidden=may_change_hidden,
        may_change_enable_object_assignments=may_change_enable_object_assignments,
    )



# coding: utf-8
"""

"""
import typing

import bleach
from markdown import markdown as _markdown_to_html
from markdown.extensions.toc import TocExtension, slugify_unicode

from .. import db
from . import errors
from .schemas.utils import data_iter
from ..models import MarkdownToHTMLCacheEntry


def markdown_to_safe_html(markdown: str, use_cache: bool = True, anchor_prefix: str = 'markdown-anchor') -> str:
    """
    Convert Markdown content to (safe) HTML.

    This first escapes all HTML content in the Markdown text and then converts
    the result to HTML. The conversion may be cached to speed up repeated use
    for the same markdown content.

    :param markdown: the Markdown content to convert
    :param use_cache: whether the cache should be used
    :param anchor_prefix: the prefix to use for anchors
    :return: the content as HTML
    """

    parameters = {
        'anchor_prefix': anchor_prefix
    }

    if use_cache:
        cache_entry = MarkdownToHTMLCacheEntry.query.filter(
            MarkdownToHTMLCacheEntry.markdown == markdown,
            MarkdownToHTMLCacheEntry.parameters['anchor_prefix'].astext == anchor_prefix
        ).first()
        if cache_entry is not None and cache_entry.html is not None:
            return cache_entry.html

    toc_extension = TocExtension(
        marker='',
        permalink=True,
        permalink_class='text-muted headerlink',
        slugify=lambda value, separator: slugify_unicode(anchor_prefix + '-' + value, separator)
    )

    html = _markdown_to_html(
        bleach.clean(markdown),
        extensions=[
            'tables',
            'sane_lists',
            'fenced_code',
            toc_extension
        ]
    )

    if use_cache:
        cache_entry = MarkdownToHTMLCacheEntry(markdown=markdown, parameters=parameters, html=html)
        db.session.add(cache_entry)
        db.session.commit()

    return html


def regenerate_cache() -> None:
    """
    Regenerate the cache for markdown_to_safe_html.
    """
    cache_entries = MarkdownToHTMLCacheEntry.query.all()
    for cache_entry in cache_entries:
        parameters = {}
        if 'anchor_prefix' in cache_entry.parameters:
            parameters['anchor_prefix'] = cache_entry.parameters['anchor_prefix']
        cache_entry.html = markdown_to_safe_html(
            cache_entry.markdown,
            use_cache=False,
            **parameters
        )
        db.session.add(cache_entry)
    db.session.commit()


def clear_cache() -> None:
    """
    Clear the cache for markdown_to_safe_html.
    """
    MarkdownToHTMLCacheEntry.query.delete()
    db.session.commit()


def get_markdown_from_object_data(data: typing.Dict[str, typing.Any]) -> typing.List[str]:
    """
    Extract Markdown text properties from object data.

    :param data: the object data to get Markdown text from
    :return: a list of all found Markdown texts
    :raise errors.ValidationError: when the contained markdown data is invalid
    """
    markdown_texts = []
    for path, property in data_iter(data, filter_property_types={'text'}):
        if isinstance(property, dict) and property.get('is_markdown'):
            if isinstance(property.get('text'), str):
                markdown_texts.append(property['text'])
            elif isinstance(property.get('text'), dict):
                for text in property['text'].values():
                    markdown_texts.append(text)
            else:
                raise errors.ValidationError('text must be str or a dictionary', [str(path_element) for path_element in path])
    return markdown_texts



# coding: utf-8
"""
Add API_ACCESS_TOKEN enum value to AuthenticationType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    return enum_value_migration('authenticationtype', 'API_ACCESS_TOKEN')



# coding: utf-8
"""

"""
import typing


class ObjectDoesNotExistError(Exception):
    pass


class FileDoesNotExistError(Exception):
    pass


class ObjectVersionDoesNotExistError(Exception):
    pass


class ObjectNotSharedError(Exception):
    pass


class ObjectNotFederatedError(Exception):
    pass


class ShareDoesNotExistError(Exception):
    pass


class ShareAlreadyExistsError(Exception):
    pass


class FileNameTooLongError(Exception):
    pass


class TooManyFilesForObjectError(Exception):
    pass


class InvalidFileStorageError(Exception):
    pass


class FederationFileNotAvailableError(Exception):
    pass


class GroupDoesNotExistError(Exception):
    pass


class GroupAlreadyExistsError(Exception):
    pass


class UserDoesNotExistError(Exception):
    pass


class UserAliasDoesNotExistError(Exception):
    pass


class UserAliasAlreadyExistsError(Exception):
    pass


class UserNotMemberOfGroupError(Exception):
    pass


class UserAlreadyMemberOfGroupError(Exception):
    pass


class InvalidGroupNameError(Exception):
    pass


class ActionDoesNotExistError(Exception):
    pass


class InstrumentDoesNotExistError(Exception):
    pass


class UserAlreadyResponsibleForInstrumentError(Exception):
    pass


class UserNotResponsibleForInstrumentError(Exception):
    pass


class UndefinedUnitError(Exception):
    pass


class MismatchedUnitError(Exception):
    pass


class ProjectAlreadyExistsError(Exception):
    pass


class InvalidProjectNameError(Exception):
    pass


class ProjectDoesNotExistError(Exception):
    pass


class UserNotMemberOfProjectError(Exception):
    pass


class UserAlreadyMemberOfProjectError(Exception):
    pass


class GroupAlreadyMemberOfProjectError(Exception):
    pass


class GroupNotMemberOfProjectError(Exception):
    pass


class NoMemberWithGrantPermissionsForProjectError(Exception):
    pass


class InvalidSubprojectRelationshipError(Exception):
    pass


class SubprojectRelationshipDoesNotExistError(Exception):
    pass


class CreatingObjectsDisabledError(Exception):
    pass


class UnauthorizedRequestError(Exception):
    pass


class RequestServerError(Exception):
    pass


class RequestError(Exception):
    pass


class ComponentNotConfiguredForFederationError(Exception):
    pass


class NoAuthenticationMethodError(Exception):
    pass


class AuthenticationMethodDoesNotExistError(Exception):
    pass


class SchemaError(Exception):
    def __init__(self, message: str, path: typing.List[str]) -> None:
        self.raw_message = message
        if path:
            message += ' (at ' + ' -> '.join(path) + ')'
        super().__init__(message)
        self.path = path
        self.message = message
        self.paths = [path]


class ValidationError(SchemaError):
    pass


class ValidationMultiError(ValidationError):
    def __init__(self, errors: typing.List[ValidationError]) -> None:
        message = '\n'.join(error.message for error in errors)
        super().__init__(message, [])
        self.paths = [error.path for error in errors]


class LocationDoesNotExistError(Exception):
    pass


class CyclicLocationError(Exception):
    pass


class ObjectLocationAssignmentDoesNotExistError(Exception):
    pass


class NotificationDoesNotExistError(Exception):
    pass


class OnlyOneAuthenticationMethod(Exception):
    pass


class AuthenticationMethodWrong(Exception):
    pass


class AuthenticationMethodAlreadyExists(Exception):
    pass


class NoEmailInLDAPAccountError(Exception):
    pass


class LDAPAccountAlreadyExistError(Exception):
    pass


class LDAPAccountOrPasswordWrongError(Exception):
    pass


class PublicationDoesNotExistError(Exception):
    pass


class InvalidDOIError(Exception):
    pass


class LDAPNotConfiguredError(Exception):
    pass


class UserIsReadonlyError(Exception):
    pass


class InstrumentLogEntryDoesNotExistError(Exception):
    pass


class InstrumentLogFileAttachmentDoesNotExistError(Exception):
    pass


class InstrumentLogObjectAttachmentDoesNotExistError(Exception):
    pass


class InstrumentLogCategoryDoesNotExistError(Exception):
    pass


class GroupInvitationDoesNotExistError(Exception):
    pass


class ProjectInvitationDoesNotExistError(Exception):
    pass


class UserInvitationDoesNotExistError(Exception):
    pass


class ActionTypeDoesNotExistError(Exception):
    pass


class TopicDoesNotExistError(Exception):
    pass


class InvalidComponentNameError(Exception):
    pass


class InvalidComponentAddressError(Exception):
    pass


class InsecureComponentAddressError(Exception):
    pass


class InvalidComponentUUIDError(Exception):
    pass


class ComponentAlreadyExistsError(Exception):
    def __init__(self, is_current_app: bool = False):
        self.is_current_app = is_current_app
        super().__init__()


class ComponentDoesNotExistError(Exception):
    pass


class ComponentValidationError(Exception):
    pass


class CommentDoesNotExistError(Exception):
    pass


class DataverseNotReachableError(Exception):
    pass


class ProjectObjectLinkDoesNotExistsError(Exception):
    pass


class ProjectObjectLinkAlreadyExistsError(Exception):
    pass


class LanguageDoesNotExistError(Exception):
    pass


class ActionTranslationDoesNotExistError(Exception):
    pass


class ActionTypeTranslationDoesNotExistError(Exception):
    pass


class InstrumentTranslationDoesNotExistError(Exception):
    pass


class MissingEnglishTranslationError(Exception):
    pass


class InvalidNumberError(Exception):
    pass


class InvalidTemplateIDError(Exception):
    pass


class RecursiveTemplateError(Exception):
    pass


class LanguageAlreadyExistsError(Exception):
    pass


class TwoFactorAuthenticationMethodDoesNotExistError(Exception):
    pass


class InvalidDataExportError(Exception):
    pass


class InvalidJSONError(Exception):
    pass


class InvalidTokenError(Exception):
    pass


class TokenExistsError(Exception):
    pass


class MissingComponentAddressError(Exception):
    pass


class BackgroundTaskDoesNotExistError(Exception):
    pass


class InvalidURLError(Exception):
    pass


class URLTooLongError(Exception):
    pass


class InvalidIPAddressError(Exception):
    pass


class InvalidPortNumberError(Exception):
    pass


class InvalidDefaultPermissionsError(Exception):
    pass


class SciCatNotReachableError(Exception):
    pass


class ObjectLocationAssignmentAlreadyConfirmedError(Exception):
    pass


class ObjectLocationAssignmentAlreadyDeclinedError(Exception):
    pass


class LocationTypeDoesNotExistError(Exception):
    pass


class GroupCategoryDoesNotExistError(Exception):
    pass


class CyclicGroupCategoryError(Exception):
    pass


class FileCreationError(Exception):
    pass


class ExceedingLocationCapacityError(Exception):
    pass


class InvalidUnitsError(Exception):
    pass


class DiffMismatchError(Exception):
    pass


class InstrumentObjectLinkAlreadyExistsError(Exception):
    pass


class ELNImportDoesNotExistError(Exception):
    pass


class InvalidELNFileError(Exception):
    pass


class WebhookAlreadyExistsError(Exception):
    pass


class WebhookDoesNotExistError(Exception):
    pass


class WebhookConnectionException(Exception):
    pass


class ObjectLogEntryDoesNotExistError(Exception):
    pass


class TemplateValueError(Exception):
    pass


class FederatedUserInFederatedIdentityError(Exception):
    pass


class ELNUserInFederatedIdentityError(Exception):
    pass


class FederatedIdentityNotFoundError(Exception):
    pass


class MismatchedResponseToRequestError(Exception):
    pass


class FailedMetadataFetchingError(Exception):
    pass


class InvalidSAMLRequestError(Exception):
    pass


class AuthnRequestParsingError(Exception):
    pass


class UnavailableArtifactError(Exception):
    pass


class IdentityProviderMetadataUnavailableError(Exception):
    pass


class ServiceProviderMetadataUnavailableError(Exception):
    pass


class FederatedLoginError(Exception):
    pass


class InfoPageDoesNotExistError(Exception):
    pass


class UpdatableObjectsCheckAlreadyInProgressError(Exception):
    pass


class UpdatableObjectsCheckDoesNotExistError(Exception):
    pass


class AutomaticSchemaUpdateAlreadyInProgressError(Exception):
    pass


class AutomaticSchemaUpdateDoesNotExistError(Exception):
    pass



# coding: utf-8
"""
Model for instrument log entries

This should not to be confused with internal logging like the object log.
"""

import enum
import datetime
import typing

from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .instruments import Instrument
from .objects import Objects
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User


instrument_log_entry_category_association_table = db.Table(
    'instrument_log_entry_version_category_associations',
    db.metadata,
    db.Column('log_entry_id', db.Integer),
    db.Column('log_entry_version_id', db.Integer),
    db.Column('category_id', db.Integer, db.ForeignKey('instrument_log_categories.id', ondelete="CASCADE")),
    db.ForeignKeyConstraint(
        ['log_entry_id', 'log_entry_version_id'],
        ['instrument_log_entry_versions.log_entry_id', 'instrument_log_entry_versions.version_id']
    )
)


@enum.unique
class InstrumentLogCategoryTheme(enum.Enum):
    GRAY = 0
    BLUE = 1
    GREEN = 2
    YELLOW = 3
    RED = 4


class InstrumentLogCategory(Model):
    __tablename__ = 'instrument_log_categories'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    instrument_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Instrument.id), nullable=False)
    title: Mapped[str] = db.Column(db.String, nullable=False)
    theme: Mapped[InstrumentLogCategoryTheme] = db.Column(db.Enum(InstrumentLogCategoryTheme), nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogCategory"]]

    def __init__(
            self,
            instrument_id: int,
            title: str,
            theme: InstrumentLogCategoryTheme
    ) -> None:
        super().__init__(
            instrument_id=instrument_id,
            title=title,
            theme=theme
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, instrument_id={self.instrument_id}, title="{self.title}", theme={self.theme.name.lower()})>'


class InstrumentLogEntry(Model):
    __tablename__ = 'instrument_log_entries'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    instrument_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Instrument.id), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False)
    author: Mapped['User'] = relationship('User')
    versions: Mapped[typing.List['InstrumentLogEntryVersion']] = relationship('InstrumentLogEntryVersion', lazy="joined")

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogEntry"]]

    def __init__(
            self,
            instrument_id: int,
            user_id: int
    ) -> None:
        super().__init__(
            instrument_id=instrument_id,
            user_id=user_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, instrument_id={self.instrument_id}, user_id={self.user_id})>'


class InstrumentLogEntryVersion(Model):
    __tablename__ = 'instrument_log_entry_versions'

    log_entry_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), primary_key=True)
    version_id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    content: Mapped[str] = db.Column(db.Text, nullable=False)
    utc_datetime: Mapped[datetime.datetime] = db.Column(db.TIMESTAMP(timezone=True), nullable=False)
    content_is_markdown: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())
    event_utc_datetime: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    categories: Mapped[typing.List[InstrumentLogCategory]] = relationship('InstrumentLogCategory', secondary=instrument_log_entry_category_association_table, lazy="joined")
    log_entry: Mapped[InstrumentLogEntry] = relationship(InstrumentLogEntry, back_populates='versions')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogEntryVersion"]]

    def __init__(
            self,
            log_entry_id: int,
            version_id: int,
            content: str,
            utc_datetime: typing.Optional[datetime.datetime] = None,
            content_is_markdown: bool = False,
            event_utc_datetime: typing.Optional[datetime.datetime] = None
    ) -> None:
        super().__init__(
            log_entry_id=log_entry_id,
            version_id=version_id,
            content=content,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            content_is_markdown=content_is_markdown,
            event_utc_datetime=event_utc_datetime
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(log_entry_id={self.log_entry_id}, version_id={self.version_id},  utc_datetime={self.utc_datetime}, content="{self.content}")>'


class InstrumentLogFileAttachment(Model):
    __tablename__ = 'instrument_log_file_attachments'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    log_entry_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), nullable=False)
    file_name: Mapped[str] = db.Column(db.String, nullable=False)
    content: Mapped[bytes] = db.Column(db.LargeBinary, nullable=False)
    is_hidden: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)
    image_info: Mapped['InstrumentLogFileAttachmentImageInfo'] = relationship('InstrumentLogFileAttachmentImageInfo', backref='file_attachment', uselist=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogFileAttachment"]]

    def __init__(
            self,
            log_entry_id: int,
            file_name: str,
            content: bytes
    ) -> None:
        super().__init__(
            log_entry_id=log_entry_id,
            file_name=file_name,
            content=content
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, log_entry_id={self.log_entry_id}, file_name="{self.file_name}")>'


class InstrumentLogFileAttachmentImageInfo(Model):
    __tablename__ = 'instrument_log_file_attachment_image_infos'

    file_attachment_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(InstrumentLogFileAttachment.id), primary_key=True)
    thumbnail_content: Mapped[bytes] = db.Column(db.LargeBinary, nullable=False)
    thumbnail_mime_type: Mapped[str] = db.Column(db.String, nullable=False)
    width: Mapped[int] = db.Column(db.Integer, nullable=False)
    height: Mapped[int] = db.Column(db.Integer, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogFileAttachmentImageInfo"]]


class InstrumentLogObjectAttachment(Model):
    __tablename__ = 'instrument_log_object_attachments'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    log_entry_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(InstrumentLogEntry.id), nullable=False)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    is_hidden: Mapped[bool] = db.Column(db.Boolean, default=False, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["InstrumentLogObjectAttachment"]]

    def __init__(
            self,
            log_entry_id: int,
            object_id: int
    ) -> None:
        super().__init__(
            log_entry_id=log_entry_id,
            object_id=object_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, log_entry_id={self.log_entry_id}, object_id={self.object_id})>'



# coding: utf-8
"""
Replace NOT NULL constraints per column by NOT NULL constraints conditioned by federation reference in actions.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    constraints = db.session.execute(db.text("""
             SELECT conname
             FROM pg_catalog.pg_constraint
             WHERE conname = 'actions_not_null_check'
        """)).fetchall()
    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE actions
            ADD CONSTRAINT actions_not_null_check
                CHECK ((
                    fed_id IS NOT NULL AND
                    component_id IS NOT NULL
                ) OR (
                    type_id IS NOT NULL AND
                    schema IS NOT NULL
                )),
            ALTER COLUMN type_id DROP NOT NULL,
            ALTER COLUMN schema DROP NOT NULL
    """))
    return True



import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_bool, _get_translation
from ..locations import get_location_type, create_location_type, update_location_type, LocationType
from ..components import Component
from .. import errors, fed_logs


class LocationTypeRef(typing.TypedDict):
    location_type_id: int
    component_uuid: str


class LocationTypeData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    name: typing.Optional[typing.Dict[str, str]]
    location_name_singular: typing.Optional[typing.Dict[str, str]]
    location_name_plural: typing.Optional[typing.Dict[str, str]]
    admin_only: bool
    enable_parent_location: bool
    enable_sub_locations: bool
    enable_object_assignments: bool
    enable_responsible_users: bool
    show_location_log: bool


class SharedLocationTypeData(typing.TypedDict):
    location_type_id: int
    component_uuid: str
    name: typing.Optional[typing.Dict[str, str]]
    location_name_singular: typing.Optional[typing.Dict[str, str]]
    location_name_plural: typing.Optional[typing.Dict[str, str]]
    admin_only: bool
    enable_parent_location: bool
    enable_sub_locations: bool
    enable_object_assignments: bool
    enable_responsible_users: bool
    show_location_log: bool


def import_location_type(
        location_type_data: LocationTypeData,
        component: Component
) -> LocationType:
    component_id = _get_or_create_component_id(location_type_data['component_uuid'])

    try:
        location_type = get_location_type(location_type_data['fed_id'], component_id)

        ignored_keys = {
            'fed_id',
            'component_uuid'
        }
        if any(
                value != getattr(location_type, key)
                for key, value in location_type_data.items()
                if key not in ignored_keys
        ):
            update_location_type(
                location_type_id=location_type.id,
                name=location_type_data['name'],
                location_name_singular=location_type_data['location_name_singular'],
                location_name_plural=location_type_data['location_name_plural'],
                admin_only=location_type_data['admin_only'],
                enable_parent_location=location_type_data['enable_parent_location'],
                enable_sub_locations=location_type_data['enable_sub_locations'],
                enable_object_assignments=location_type_data['enable_object_assignments'],
                enable_responsible_users=location_type_data['enable_responsible_users'],
                enable_instruments=False,
                enable_capacities=False,
                show_location_log=location_type_data['show_location_log'],
            )
            fed_logs.update_location_type(location_type.id, component.id)
    except errors.LocationTypeDoesNotExistError:
        location_type = create_location_type(
            name=location_type_data['name'],
            location_name_singular=location_type_data['location_name_singular'],
            location_name_plural=location_type_data['location_name_plural'],
            admin_only=location_type_data['admin_only'],
            enable_parent_location=location_type_data['enable_parent_location'],
            enable_sub_locations=location_type_data['enable_sub_locations'],
            enable_object_assignments=location_type_data['enable_object_assignments'],
            enable_responsible_users=location_type_data['enable_responsible_users'],
            enable_instruments=False,
            enable_capacities=False,
            show_location_log=location_type_data['show_location_log'],
            fed_id=location_type_data['fed_id'],
            component_id=component_id,
        )
        fed_logs.import_location_type(location_type.id, component.id)
    return location_type


def parse_import_location_type(
        location_type_data: typing.Dict[str, typing.Any],
        component: Component
) -> LocationType:
    return import_location_type(parse_location_type(location_type_data), component)


def parse_location_type(
        location_type_data: typing.Dict[str, typing.Any]
) -> LocationTypeData:
    fed_id = _get_id(location_type_data.get('location_type_id'), special_values=[LocationType.LOCATION])
    uuid = _get_uuid(location_type_data.get('component_uuid'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local location type {fed_id}')
    return LocationTypeData(
        fed_id=fed_id,
        component_uuid=uuid,
        name=_get_translation(location_type_data.get('name')),
        location_name_singular=_get_translation(location_type_data.get('location_name_singular')),
        location_name_plural=_get_translation(location_type_data.get('location_name_plural')),
        admin_only=_get_bool(location_type_data.get('admin_only'), default=True),
        enable_parent_location=_get_bool(location_type_data.get('enable_parent_location'), default=False),
        enable_sub_locations=_get_bool(location_type_data.get('enable_sub_locations'), default=False),
        enable_object_assignments=_get_bool(location_type_data.get('enable_object_assignments'), default=False),
        enable_responsible_users=_get_bool(location_type_data.get('enable_responsible_users'), default=False),
        show_location_log=_get_bool(location_type_data.get('show_location_log'), default=False),
    )


def _parse_location_type_ref(
        location_type_data: typing.Optional[typing.Union[LocationTypeRef, typing.Dict[str, typing.Any]]]
) -> typing.Optional[LocationTypeRef]:
    if location_type_data is None:
        return None
    location_type_id = _get_id(location_type_data.get('location_type_id'), special_values=[LocationType.LOCATION])
    component_uuid = _get_uuid(location_type_data.get('component_uuid'))
    return LocationTypeRef(
        location_type_id=location_type_id,
        component_uuid=component_uuid
    )


@typing.overload
def _get_or_create_location_type_id(
        location_type_data: LocationTypeRef
) -> int:
    ...


@typing.overload
def _get_or_create_location_type_id(
        location_type_data: None
) -> None:
    ...


def _get_or_create_location_type_id(
        location_type_data: typing.Optional[LocationTypeRef]
) -> typing.Optional[int]:
    if location_type_data is None:
        return None
    component_id = _get_or_create_component_id(location_type_data['component_uuid'])
    try:
        location_type = get_location_type(location_type_data['location_type_id'], component_id)
    except errors.LocationTypeDoesNotExistError:
        assert component_id is not None
        location_type = create_location_type(
            name=None,
            location_name_singular=None,
            location_name_plural=None,
            admin_only=True,
            enable_sub_locations=False,
            enable_parent_location=False,
            enable_object_assignments=False,
            enable_responsible_users=False,
            enable_instruments=False,
            enable_capacities=False,
            show_location_log=False,
            fed_id=location_type_data['location_type_id'],
            component_id=component_id
        )
        fed_logs.create_ref_location_type(location_type.id, component_id)
    return location_type.id


def shared_location_type_preprocessor(
        location_type_id: int,
        _component: Component,
        _refs: typing.List[typing.Tuple[str, int]],
        _markdown_images: typing.Dict[str, str]
) -> typing.Optional[SharedLocationTypeData]:
    location_type = get_location_type(location_type_id)
    if location_type.component_id is not None:
        return None
    return SharedLocationTypeData(
        location_type_id=location_type.id if location_type.fed_id is None else location_type.fed_id,
        component_uuid=flask.current_app.config['FEDERATION_UUID'] if location_type.component is None else location_type.component.uuid,
        name=location_type.name,
        location_name_singular=location_type.location_name_singular,
        location_name_plural=location_type.location_name_plural,
        admin_only=location_type.admin_only,
        enable_parent_location=location_type.enable_parent_location,
        enable_sub_locations=location_type.enable_sub_locations,
        enable_object_assignments=location_type.enable_object_assignments,
        enable_responsible_users=location_type.enable_responsible_users,
        show_location_log=location_type.show_location_log,
    )



# coding: utf-8
"""
Script for disabling two-factor authentication for a user.

Usage: sampledb disable_two_factor_authentication <user_id>
"""

import sys
import typing

from .. import create_app
from ..logic.authentication import get_active_two_factor_authentication_methods, deactivate_two_factor_authentication_method


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 1 or not all(arguments):
        print(__doc__)
        sys.exit(1)
    try:
        user_id = int(arguments[0])
    except ValueError:
        print("Error: user_id must be a valid user_id", file=sys.stderr)
        sys.exit(1)
    app = create_app()
    with app.app_context():
        active_methods = get_active_two_factor_authentication_methods(user_id)
        if not active_methods:
            print("Error: the user does not have two-factor authentication enabled", file=sys.stderr)
            sys.exit(1)
        for active_method in active_methods:
            deactivate_two_factor_authentication_method(active_method.id)
        print("Success: two-factor authentication has been disabled for the user")



# coding: utf-8
"""

"""

import flask
import flask_login
from flask_babel import _

from .. import frontend
from ... import logic
from ...logic.actions import get_action
from ...logic.action_types import get_action_type
from ...logic.object_permissions import get_user_object_permissions
from ...logic.settings import get_user_setting
from ...logic.objects import get_object, get_object_versions, get_current_object_version_id
from ...logic.languages import get_languages_in_object_data, get_language, get_languages, Language
from ...logic.errors import ObjectDoesNotExistError, ValidationError
from ...logic.components import get_component
from ...logic.utils import get_translated_text
from ...logic.schemas.data_diffs import calculate_diff
from .forms import ObjectVersionRestoreForm
from ...utils import object_permissions_required, FlaskResponseT
from ..utils import get_user_if_exists
from .permissions import on_unauthorized, get_object_if_current_user_has_read_permissions
from ...models import Permissions


@frontend.route('/objects/<int:object_id>/versions/')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_versions(object_id: int) -> FlaskResponseT:
    object = get_object(object_id=object_id)
    if object is None:
        return flask.abort(404)
    object_versions = get_object_versions(object_id=object_id)
    object_versions.sort(key=lambda object_version: -object_version.version_id)
    return flask.render_template('objects/object_versions.html', get_user=get_user_if_exists, object=object, object_versions=object_versions)


@frontend.route('/objects/<int:object_id>/versions/<int:version_id>')
@object_permissions_required(Permissions.READ, on_unauthorized=on_unauthorized)
def object_version(object_id: int, version_id: int) -> FlaskResponseT:
    object = get_object(object_id=object_id, version_id=version_id)
    previous_version_data_diff = None
    previous_version_schema = None
    if 'diff' in flask.request.args and object.data is not None:
        if version_id > 0:
            try:
                previous_version_object = get_object(object_id=object_id, version_id=version_id - 1)
            except logic.errors.ObjectVersionDoesNotExistError:
                previous_version_object = None
            if previous_version_object is not None and previous_version_object.data is not None and previous_version_object.schema is not None:
                previous_version_data_diff = calculate_diff(previous_version_object.data, object.data)
                previous_version_schema = previous_version_object.schema
    current_version_id = get_current_object_version_id(object_id=object_id)
    form = None
    user_permissions = get_user_object_permissions(object_id=object_id, user_id=flask_login.current_user.id)
    if Permissions.WRITE in user_permissions:
        current_object = get_object(object_id=object_id)
        if current_object.version_id != version_id:
            form = ObjectVersionRestoreForm()
    user_may_grant = Permissions.GRANT in user_permissions
    action = get_action(object.action_id) if object.action_id else None
    action_type = get_action_type(action.type_id) if action and action.type_id else None
    instrument = action.instrument if action else None

    # languages
    english = get_language(Language.ENGLISH)
    all_languages = get_languages()
    languages_by_lang_code = {
        language.lang_code: language
        for language in all_languages
    }
    if object.data:
        object_languages = [
            languages_by_lang_code.get(lang_code)
            for lang_code in get_languages_in_object_data(object.data)
        ]
    else:
        object_languages = []
    metadata_language = flask.request.args.get('language', None)
    if metadata_language not in languages_by_lang_code:
        metadata_language = None

    return flask.render_template(
        'objects/view/base.html',
        template_mode="view",
        show_object_type_and_id_on_object_page_text=get_user_setting(flask_login.current_user.id, "SHOW_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE"),
        show_object_title=get_user_setting(flask_login.current_user.id, "SHOW_OBJECT_TITLE"),
        languages=object_languages,
        object_languages=object_languages,
        metadata_language=metadata_language,
        ENGLISH=english,
        is_archived=True,
        object=object,
        object_type=get_translated_text(action_type.object_name) if action_type is not None else None,
        action=action,
        action_type=action_type,
        instrument=instrument,
        schema=object.schema,
        data=object.data,
        name=object.name,
        get_object_if_current_user_has_read_permissions=get_object_if_current_user_has_read_permissions,
        object_id=object_id,
        version_id=version_id,
        current_version_id=current_version_id,
        previous_version_data_diff=previous_version_data_diff,
        previous_version_schema=previous_version_schema,
        link_version_specific_rdf=True,
        restore_form=form,
        get_user=get_user_if_exists,
        user_may_grant=user_may_grant,
        get_action_type=get_action_type,
        component=object.component,
        fed_object_id=object.fed_object_id,
        fed_version_id=object.fed_version_id,
        get_component=get_component
    )


@frontend.route('/objects/<int:object_id>/versions/<int:version_id>/restore', methods=['GET', 'POST'])
@object_permissions_required(Permissions.WRITE)
def restore_object_version(object_id: int, version_id: int) -> FlaskResponseT:
    if version_id < 0 or object_id < 0:
        return flask.abort(404)
    try:
        current_object = get_object(object_id=object_id)
    except ObjectDoesNotExistError:
        return flask.abort(404)
    if current_object.version_id <= version_id:
        return flask.abort(404)
    form = ObjectVersionRestoreForm()
    if form.validate_on_submit():
        try:
            logic.objects.restore_object_version(object_id=object_id, version_id=version_id, user_id=flask_login.current_user.id)
        except ValidationError:
            flask.flash(_('This version contains invalid data and cannot be restored.'), 'error')
        else:
            return flask.redirect(flask.url_for('.object', object_id=object_id))
    return flask.render_template('objects/restore_object_version.html', object_id=object_id, version_id=version_id, restore_form=form)



# coding: utf-8
"""
Replace string columns with JSON columns in projects table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    column_names = db.session.execute(db.text("""
        SELECT column_name, data_type
        FROM information_schema.columns
        WHERE table_name = 'projects'
    """)).fetchall()
    if ('name', 'json') in column_names and ('description', 'json') in column_names:
        return False

    # Perform migration
    existing_data = [
        project_data
        for project_data in db.session.execute(db.text("""
        SELECT id, name, description
        FROM projects
        """))
    ]

    db.session.execute(db.text("""
        ALTER TABLE projects
        DROP COLUMN name;
    """))

    db.session.execute(db.text("""
        ALTER TABLE projects
        DROP COLUMN description;
    """))

    db.session.execute(db.text("""
          ALTER TABLE projects
          ADD name JSON NOT NULL DEFAULT '{}'::json;
      """))
    db.session.execute(db.text("""
          ALTER TABLE projects
          ADD description JSON NOT NULL DEFAULT '{}'::json;
      """))

    for id, name, description in existing_data:
        project_data = {
            'id': id,
            'name': name,
            'description': description
        }
        db.session.execute(db.text("""
            UPDATE projects
            SET name = json_build_object('en', :name), description = json_build_object('en', :description)
            WHERE id =:id
        """), params=project_data)

    return True




from http import HTTPStatus

import flask
import flask_login

from . import frontend
from ..logic.background_tasks import get_background_tasks
from ..utils import FlaskResponseT


@frontend.route('/admin/background_tasks/')
@flask_login.login_required
def background_tasks() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(HTTPStatus.FORBIDDEN)
    return flask.render_template(
        'admin/background_tasks.html',
        tasks=get_background_tasks()
    )



# coding: utf-8
"""
Logic module for management of languages
"""

import dataclasses
import typing

from .caching import cache_per_request
from .. import db
from . import errors, settings, locale
from .. import models
if typing.TYPE_CHECKING:
    from .users import User


@dataclasses.dataclass(frozen=True)
class Language:
    """
    This class provides an immutable wrapper around models.languages.Language.
    """
    id: int
    lang_code: str
    names: typing.Dict[str, str]
    datetime_format_datetime: str
    datetime_format_moment: str
    datetime_format_moment_output: str
    date_format_moment_output: str
    enabled_for_input: bool
    enabled_for_user_interface: bool

    ENGLISH = models.Language.ENGLISH
    GERMAN = models.Language.GERMAN

    @staticmethod
    def from_database(language: models.Language) -> 'Language':
        return Language(
            id=language.id,
            lang_code=language.lang_code,
            names=language.names,
            datetime_format_datetime=language.datetime_format_datetime,
            datetime_format_moment=language.datetime_format_moment,
            datetime_format_moment_output=language.datetime_format_moment_output,
            date_format_moment_output=language.date_format_moment_output,
            enabled_for_input=language.enabled_for_input,
            enabled_for_user_interface=language.enabled_for_user_interface
        )


def create_language(
        *,
        names: typing.Dict[str, str],
        lang_code: str,
        datetime_format_datetime: str,
        datetime_format_moment: str,
        datetime_format_moment_output: str,
        date_format_moment_output: str,
        enabled_for_input: bool,
        enabled_for_user_interface: bool
) -> Language:
    """
    Create a new language.

    :param names: a dict containing the language's names in other languages
    :param lang_code: the language code
    :param datetime_format_datetime: format for datetime
    :param datetime_format_moment: format for moment
    :param datetime_format_moment_output: output format for moment
    :param date_format_moment_output: output format for moment for dates only
    :param enabled_for_input: whether the language is enabled for input
    :param enabled_for_user_interface: whether the language is enabled
        for the user interface
    :return: the new language
    :raise errors.LanguageAlreadyExistsError: if another language with this
        language code already exists
    :raise errors.LanguageDoesNotExistError: if the language name contains a
        translation for an unknown language code
    """
    if models.Language.query.filter_by(lang_code=lang_code).first() is not None:
        raise errors.LanguageAlreadyExistsError()

    all_language_codes = {
        language.lang_code
        for language in models.Language.query.all()
    }
    all_language_codes.add(lang_code)

    if set(names.keys()) - all_language_codes:
        raise errors.LanguageDoesNotExistError()

    language = models.Language(
        names=names,
        lang_code=lang_code,
        datetime_format_datetime=datetime_format_datetime,
        datetime_format_moment=datetime_format_moment,
        datetime_format_moment_output=datetime_format_moment_output,
        date_format_moment_output=date_format_moment_output,
        enabled_for_input=enabled_for_input,
        enabled_for_user_interface=enabled_for_user_interface
    )
    db.session.add(language)
    db.session.commit()
    return Language.from_database(language)


def update_language(
        language_id: int,
        names: typing.Dict[str, str],
        lang_code: str,
        datetime_format_datetime: str,
        datetime_format_moment: str,
        datetime_format_moment_output: str,
        date_format_moment_output: str,
        enabled_for_input: bool,
        enabled_for_user_interface: bool
) -> None:
    """
    Update a language.

    :param language_id: the id of the language
    :param names: a dict containing the language's names in other languages
    :param lang_code: the language code
    :param datetime_format_datetime: format for datetime
    :param datetime_format_moment: format for moment
    :param datetime_format_moment_output: format for moment output
    :param date_format_moment_output: output format for moment for dates only
    :param enabled_for_input: whether the language is enabled for input
    :param enabled_for_user_interface: whether the language is enabled
        for the user interface
    :raise errors.LanguageAlreadyExistsError: when the language code already
        exists for a different language or the current language code matches
        a supported locale language code
    :raise errors.LanguageDoesNotExistError: if the language name contains a
        translation for an unknown language code or the language itself does
        not exist
    """

    all_language_codes = {
        language.lang_code
        for language in models.Language.query.all()
    }
    all_language_codes.add(lang_code)

    language = models.Language.query.filter_by(id=language_id).first()
    if language is None:
        raise errors.LanguageDoesNotExistError()
    if language.lang_code != lang_code:
        if models.Language.query.filter_by(lang_code=lang_code).first() is not None:
            raise errors.LanguageAlreadyExistsError()
        if language.lang_code in locale.SUPPORTED_LOCALES:
            raise errors.LanguageAlreadyExistsError()
        all_language_codes.remove(language.lang_code)

    if set(names.keys()) - all_language_codes:
        raise errors.LanguageDoesNotExistError()

    language.names = names
    language.lang_code = lang_code
    language.datetime_format_datetime = datetime_format_datetime
    language.datetime_format_moment = datetime_format_moment
    language.datetime_format_moment_output = datetime_format_moment_output
    language.date_format_moment_output = date_format_moment_output
    language.enabled_for_input = enabled_for_input
    language.enabled_for_user_interface = enabled_for_user_interface
    db.session.add(language)
    db.session.commit()


def get_languages(
        only_enabled_for_input: bool = False
) -> typing.List[Language]:
    """
    Return all languages.

    :param only_enabled_for_input: if True, only return languages which are
        enabled for input

    :return: the list of all languages
    """
    language_query = models.Language.query
    if only_enabled_for_input:
        language_query = language_query.filter_by(enabled_for_input=True)
    language_query = language_query.order_by(models.Language.id)
    return [
        Language.from_database(language)
        for language in language_query.all()
    ]


def get_language(language_id: int) -> Language:
    """
    Returns the language with the given language ID.

    :param language_id: the ID of an existing language
    :return: the language
    :raise errors.LanguageDoesNotExistError: when no language with the given language ID exists
    """
    language = models.Language.query.filter_by(id=language_id).first()
    if language is None:
        raise errors.LanguageDoesNotExistError()
    return Language.from_database(language)


def get_language_by_lang_code(lang_code: str) -> Language:
    """
    Returns the language with the given lang_code.

    :param lang_code: the lang_code of an existing language
    :return: the language with the lang_code
    :raise errors.LanguageDoesNotExistError: when there is no language for the given lang code
    """
    language = models.Language.query.filter_by(lang_code=lang_code).first()

    if language is None:
        raise errors.LanguageDoesNotExistError()
    return Language.from_database(language)


@cache_per_request(key=lambda user: user.id if user else None)
def get_user_language(user: typing.Optional['User']) -> Language:
    """
    Return the language of the current user.

    :param user: the current user, or None
    :return: the user's language or english
    """
    if not user or not user.is_authenticated:
        try:
            return get_language_by_lang_code(locale.guess_request_locale())
        except errors.LanguageDoesNotExistError:
            return get_language(models.Language.ENGLISH)

    language = user.language_cache[0]
    if language is None:
        auto_lc = settings.get_user_setting(user.id, 'AUTO_LC')
        if auto_lc:
            language_code = locale.guess_request_locale()
        else:
            language_code = settings.get_user_setting(user.id, 'LOCALE')
        try:
            language = get_language_by_lang_code(language_code)
        except errors.LanguageDoesNotExistError:
            language = get_language(models.Language.ENGLISH)
        user.language_cache[0] = language
    return language


def get_languages_in_object_data(
        data: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]
) -> typing.Set[str]:
    language_codes = set()

    if isinstance(data, dict):
        if '_type' in data:
            if data['_type'] == 'text' and isinstance(data.get('text'), dict):
                language_codes.update(set(data['text']))
        else:
            for property_data in data.values():
                language_codes.update(get_languages_in_object_data(property_data))

    if isinstance(data, list):
        for item_data in data:
            language_codes.update(get_languages_in_object_data(item_data))

    return language_codes


def filter_translations(
        translations: typing.Dict[str, str]
) -> typing.Dict[str, str]:
    allowed_language_codes = {
        language.lang_code
        for language in get_languages(only_enabled_for_input=True)
    }

    filtered_translations = {}
    for language_code, translation in translations.items():
        if language_code not in allowed_language_codes:
            raise errors.LanguageDoesNotExistError()
        if translation:
            filtered_translations[language_code] = translation

    return filtered_translations


def get_language_codes(
        only_enabled_for_input: bool = False,
        only_enabled_for_user_interface: bool = False
) -> typing.Set[str]:
    """
    Return a set of known language codes.

    :param only_enabled_for_input: only return codes for languages enabled
        for input
    :param only_enabled_for_user_interface: only return codes for languages
        enabled for the user interface
    :return: the set of language codes
    """
    query = models.Language.query
    if only_enabled_for_input:
        query = query.filter_by(enabled_for_input=True)
    if only_enabled_for_user_interface:
        query = query.filter_by(enabled_for_user_interface=True)
    language_code_tuples = query.with_entities(models.Language.lang_code).all()
    return {
        language_code_tuple[0]
        for language_code_tuple in language_code_tuples
    }



# coding: utf-8
"""

"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .groups import Group
from .users import User
from .locations import Location
from .permissions import Permissions
from .utils import Model

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


class UserLocationPermissions(Model):
    __tablename__ = 'user_location_permissions'

    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Location.id), nullable=False)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["UserLocationPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(location_id, user_id),
    )


class GroupLocationPermissions(Model):
    __tablename__ = 'group_location_permissions'

    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Location.id), nullable=False)
    group_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Group.id, ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["GroupLocationPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(location_id, group_id),
    )


class ProjectLocationPermissions(Model):
    __tablename__ = 'project_location_permissions'

    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Location.id), nullable=False)
    project_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey('projects.id', ondelete="CASCADE"), nullable=False)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ProjectLocationPermissions"]]

    __table_args__ = (
        db.PrimaryKeyConstraint(location_id, project_id),
    )


class AllUserLocationPermissions(Model):
    __tablename__ = 'all_user_location_permissions'

    location_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Location.id), primary_key=True)
    permissions: Mapped[Permissions] = db.Column(db.Enum(Permissions), nullable=False, default=Permissions.NONE)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["AllUserLocationPermissions"]]



# coding: utf-8
"""

"""

import base64
import functools
import json
import os
import typing
import secrets

import flask
import flask_login
import sqlalchemy
import werkzeug

from . import logic, db
from .models import Permissions, migrations

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


FlaskResponseValueT = typing.Union[flask.Response, werkzeug.Response, str]
FlaskResponseT = typing.Union[FlaskResponseValueT, typing.Tuple[FlaskResponseValueT, int], typing.Tuple[FlaskResponseValueT, int, typing.Dict[str, str]]]


def ansi_color(text: str, color: int) -> str:
    """
    Add ANSI color codes to text.

    :param text: the text without ANSI colors
    :param color: the desired color
    :return: the text with the color added
    """
    return f'\033[{color}m{text}\033[0m'


def object_permissions_required(
        required_object_permissions: Permissions,
        auth_extension: typing.Any = flask_login,
        user_id_callable: typing.Callable[[], typing.Optional[int]] = lambda: flask_login.current_user.get_id() if flask_login.current_user else None,
        on_unauthorized: typing.Callable[[int], FlaskResponseT] = lambda object_id: flask.abort(403),
        may_enable_anonymous_users: bool = True
) -> typing.Callable[[typing.Any], typing.Any]:
    def decorator(
            func: typing.Callable[[typing.Any], typing.Any],
            user_id_callable: typing.Callable[[], typing.Optional[int]] = user_id_callable,
            on_unauthorized: typing.Callable[[int], FlaskResponseT] = on_unauthorized
    ) -> typing.Callable[[typing.Any], typing.Any]:
        @functools.wraps(func)
        def wrapper(
                *args: typing.Any,
                user_id_callable: typing.Callable[[], typing.Optional[int]] = user_id_callable,
                on_unauthorized: typing.Callable[[int], FlaskResponseT] = on_unauthorized,
                **kwargs: typing.Any
        ) -> typing.Any:
            assert 'object_id' in kwargs
            object_id = kwargs['object_id']
            version_id = kwargs.get('version_id')
            try:
                if version_id is None:
                    logic.objects.check_object_exists(object_id)
                else:
                    logic.objects.check_object_version_exists(object_id, version_id)
            except logic.errors.ObjectDoesNotExistError:
                return flask.abort(404)
            except logic.errors.ObjectVersionDoesNotExistError:
                return flask.abort(404)
            user_id = user_id_callable()
            if user_id is None:
                if may_enable_anonymous_users and flask.current_app.config['ENABLE_ANONYMOUS_USERS']:
                    anonymous_permissions = logic.object_permissions.get_object_permissions_for_anonymous_users(object_id)
                else:
                    anonymous_permissions = Permissions.NONE
                if required_object_permissions not in anonymous_permissions:
                    return auth_extension.login_required(lambda: on_unauthorized(object_id))()
            elif required_object_permissions not in logic.object_permissions.get_object_permissions_for_all_users(object_id):
                user = logic.users.get_user(user_id)
                if user.is_readonly and required_object_permissions not in Permissions.READ:
                    return on_unauthorized(object_id)
                user_object_permissions = logic.object_permissions.get_user_object_permissions(object_id=object_id, user_id=user_id)
                if required_object_permissions not in user_object_permissions:
                    return on_unauthorized(object_id)
            return func(*args, **kwargs)
        if not may_enable_anonymous_users:
            wrapper = auth_extension.login_required(wrapper)
        return wrapper
    return decorator


def load_environment_configuration(env_prefix: str) -> typing.Dict[str, typing.Any]:
    """
    Loads configuration data from environment variables with a given prefix.
    If the prefixed environment variable B64_JSON_ENV exists, its content
    will be treated as an Base64 encoded JSON object and it's attributes
    starting with the prefix will be added to the environment.

    :return: a dict containing the configuration values
    """
    b64_json_env = os.environ.get(env_prefix + 'B64_JSON_ENV', None)
    if b64_json_env:
        for key, value in json.loads(base64.b64decode(b64_json_env.encode('utf-8')).decode('utf-8')).items():
            if key.startswith(env_prefix):
                os.environ[key] = value
    config = {
        key[len(env_prefix):]: value
        for key, value in os.environ.items()
        if key.startswith(env_prefix) and key != env_prefix + 'B64_JSON_ENV'
    }
    return config


def generate_secret_key(num_bits: int) -> str:
    """
    Generates a secure, random key for the application.

    :param num_bits: number of bits of random data in the secret key
    :return: the base64 encoded secret key
    """
    num_bytes = num_bits // 8
    binary_key = os.urandom(num_bytes)
    base64_key = base64.b64encode(binary_key).decode('ascii')
    return base64_key


def empty_database(
        engine: sqlalchemy.engine.Engine,
        recreate: bool = False,
        only_delete: bool = True
) -> None:
    metadata = sqlalchemy.MetaData()
    with engine.begin() as connection:
        # delete views, as SQLAlchemy cannot reflect them
        connection.execute(db.text("DROP VIEW IF EXISTS user_object_permissions_by_all"))
        # delete instruments -> objects_current foreign key, as it prevents sorting the tables
        connection.execute(db.text("ALTER TABLE IF EXISTS instruments DROP CONSTRAINT IF EXISTS instruments_object_id_fkey"))
        # delete users -> eln_imports foreign key, as it prevents sorting the tables
        connection.execute(db.text("ALTER TABLE IF EXISTS users DROP CONSTRAINT IF EXISTS fk_users_eln_import_id"))
    # delete tables, etc
    metadata.reflect(bind=engine)
    if only_delete:
        with engine.begin() as connection:
            for table in reversed(metadata.sorted_tables):
                connection.execute(table.delete())
            # migration_index needs to be dropped so migration #0 will run
            connection.execute(db.text("DROP TABLE IF EXISTS migration_index"))
    else:
        metadata.drop_all(bind=engine)
    if recreate:
        # recreate the tables
        db.metadata.create_all(bind=engine)
        # run migrations
        migrations.run(db)


def generate_content_security_policy_nonce() -> str:
    nonce = getattr(flask.g, 'content_security_policy_nonce', None)
    if nonce is None:
        nonce = secrets.token_hex(32)
        flask.g.content_security_policy_nonce = nonce
    return nonce


def text_to_bool(text: str) -> bool:
    return text.lower() not in {'', 'false', 'no', 'off', '0'}



# coding: utf-8
"""

"""
import typing

import flask
import flask_login

from .. import frontend
from ...logic.components import get_component
from ...logic.users import get_users
from ...utils import FlaskResponseT


@frontend.route('/users/')
@flask_login.login_required
def users() -> FlaskResponseT:
    return flask.render_template(
        'users.html',
        users=sorted(get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN']), key=lambda u: typing.cast(int, u.id)),
        get_component=get_component
    )
